from nipype.interfaces.base import BaseInterfaceInputSpec, BaseInterface, File, TraitedSpec, traits
from nipype import Node
from panpipelines.utils.util_functions import *
import os
import glob
import nibabel as nb
from bids import BIDSLayout
import shlex
import subprocess
from nipype import logging as nlogging

IFLOGGER=nlogging.getLogger('nipype.interface')

def fslanat_proc(labels_dict,bids_dir=""):

    cwd=os.getcwd()
    labels_dict = updateParams(labels_dict,"CWD",cwd)

    command_base, container = getContainer(labels_dict,nodename="fslanat", SPECIFIC="FSL_CONTAINER",LOGGER=IFLOGGER)
    IFLOGGER.info("Checking the fsl version:")
    command = f"{command_base} fslversion"
    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    participant_label = getParams(labels_dict,'PARTICIPANT_LABEL')
    layout = BIDSLayout(bids_dir)
    T1w=layout.get(subject=participant_label,suffix='T1w', extension='nii.gz')
    if T1w:
        structin=T1w[0].path
        structout=os.path.join(cwd,"{}_struct".format(participant_label))

    params = "--nocrop "\
             " -i " + structin + \
             " -o " + structout

    command=f"{command_base} fsl_anat"\
            " "+params

    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    fslanat_dir = structout + ".anat"
    T1w_biascorr = getGlob(os.path.join(fslanat_dir,"T1_biascorr.nii.gz"))
    T1w_biascorr_brain = getGlob(os.path.join(fslanat_dir,"T1_biascorr_brain.nii.gz"))
    T1w_biascorr_brain_mask = getGlob(os.path.join(fslanat_dir,"T1_biascorr_brain_mask.nii.gz"))
    out_files=[]
    out_files.insert(0,T1w_biascorr)
    out_files.insert(1,T1w_biascorr_brain)
    out_files.insert(2,T1w_biascorr_brain_mask)


    return {
        "T1w_biascorr":T1w_biascorr,
        "T1w_biascorr_brain":T1w_biascorr_brain,
        "T1w_biascorr_brain_mask":T1w_biascorr_brain_mask,
        "fslanat_dir":fslanat_dir,
        "out_files":out_files
    }


class fslanatInputSpec(BaseInterfaceInputSpec):
    labels_dict = traits.Dict({},mandatory=False,desc='labels', usedefault=True)
    bids_dir = traits.String("",desc="BIDS Directory", usedefault=True)

class fslanatOutputSpec(TraitedSpec):
    T1w_biascorr = File(desc='T1w_biascorr')
    T1w_biascorr_brain = File(desc='T1w_biascorr_brain')
    T1w_biascorr_brain_mask = File(desc='T1w_biascorr_brain_mask')
    fslanat_dir = traits.String(desc="output directory of fsl_anat output")
    out_files = traits.List(desc='list of files')
    
class fslanat_pan(BaseInterface):
    input_spec = fslanatInputSpec
    output_spec = fslanatOutputSpec

    def _run_interface(self, runtime):

        # Call our python code here:
        outputs = fslanat_proc(
            self.inputs.labels_dict,
            self.inputs.bids_dir
        )

        setattr(self, "_results", outputs)
        # And we are done
        return runtime

    def _list_outputs(self):
        return self._results


def create(labels_dict,name="fslanat_node",bids_dir="",LOGGER=IFLOGGER):
    # Create Node
    pan_node = Node(fslanat_pan(), name=name)

    if LOGGER:
        LOGGER.info(f"Created Node {pan_node!r}")
        
    # Specify node inputs
    pan_node.inputs.labels_dict = labels_dict

    if bids_dir is None or bids_dir == "":
        bids_dir = substitute_labels("<BIDS_DIR>", labels_dict)

    pan_node.inputs.bids_dir =  bids_dir

    return pan_node





from nipype.interfaces.base import BaseInterfaceInputSpec, BaseInterface, File, TraitedSpec, traits
from nipype import Node
from panpipelines.utils.util_functions import *
import os
import glob
import numpy as np 
import nibabel as nib
from nipype import logging as nlogging

IFLOGGER=nlogging.getLogger('nipype.interface')

def collate_csv_single_proc(labels_dict, csv_list1,csv_list2, add_prefix):

    if csv_list1 is None:
        csv_list1 = []
    if csv_list2 is None:
        csv_list2 = []

    cwd=os.getcwd()
    labels_dict = updateParams(labels_dict,"CWD",cwd)
    
    output_dir=cwd
    participant_label = getParams(labels_dict,'PARTICIPANT_LABEL')
    pipeline = getParams(labels_dict,'PIPELINE')
    session_label = getParams(labels_dict,'PARTICIPANT_SESSION')

    if not session_label:
        roi_output_dir = os.path.join(cwd,f"{participant_label}_roi_output_dir")
    else:
        roi_output_dir = os.path.join(cwd,f"{participant_label}_{session_label}_roi_output_dir")

    if not os.path.isdir(roi_output_dir):
        os.makedirs(roi_output_dir)

    csv_list=[]
    if csv_list1 is not None:
        csv_list.extend(csv_list1)

    if csv_list2 is not None:
        csv_list.extend(csv_list2)

    csv_list.sort()
    IFLOGGER.info(f"List of csv files to collate: {csv_list}")

    out_files=[]
    roi_csv = None
    if len(csv_list) > 0:
        cum_table_data = []
        cum_table_columns=[]
        for csv_file in csv_list:
            filenames = os.path.basename(csv_file).split("_")
            if len(filenames) > 3:
                prefix= filenames[1]+"_"+filenames[2]+"."
            elif len(filenames) > 2:
                prefix= filenames[1]+"."
            else:
                prefix= filenames[0]+"."

            if not add_prefix:
                prefix=""

            df = pd.read_table(csv_file,sep=",")
            if "subject_id" in df.columns:
                df = df.drop("subject_id",axis=1)
            if "session_id" in df.columns:
                df = df.drop("session_id",axis=1)
            table_columns = df.columns.tolist()
            table_columns = [prefix+x for x in table_columns]
            cum_table_columns.extend(table_columns)
            cum_table_data.extend(df.values.tolist()[0])

        cum_df = pd.DataFrame([cum_table_data])
        cum_df.columns = cum_table_columns
        
        # remove duplicates - this is definitely not the most efficient way - so keep an eye on this if dataframes get large
        # https://stackoverflow.com/questions/14984119/python-pandas-remove-duplicate-columns
        cum_df_unique = cum_df.loc[:,~cum_df.columns.duplicated()].copy()
        # sort the columns by name alphabetically
        cum_df_unique = cum_df_unique.reindex(sorted(cum_df_unique.columns), axis=1)
        if session_label is not None and not session_label == "":
            cum_df_unique.insert(0,"session_id",["ses-"+session_label])
        if participant_label is not None and not participant_label == "":
            cum_df_unique.insert(0,"subject_id",["sub-"+participant_label])
    
        collate_name = getParams(labels_dict,"COLLATE_NAME")
        if not collate_name:
            if not pipeline:
                collate_name="csvgroup"
            else:
                collate_name="pipeline"
        
        if not session_label:
            roi_csv = os.path.join(roi_output_dir,f"{participant_label}_{collate_name}.csv")
        else:
            roi_csv = os.path.join(roi_output_dir,f"{participant_label}_{session_label}_{collate_name}.csv")
        cum_df_unique.to_csv(roi_csv,sep=",",header=True, index=False)

        metadata = {}
        roi_csv_json = os.path.splitext(roi_csv)[0] + ".json"
        metadata = updateParams(metadata,"Title","collate_csv_single.py")
        metadata = updateParams(metadata,"Description","Combine csv files of participant into 1 table.")
        metadata = updateParams(metadata,"MetadataFile",f"{roi_csv_json}")
        metadata = updateParams(metadata,"FileCreated",f"{roi_csv}")
        metadata = updateParams(metadata,"DateCreated",datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S:%f"))
        metadata = updateParams(metadata,"Pipeline",f"{pipeline}")
        metadata = updateParams(metadata,"InputFiles",f"{csv_list}")
        export_labels(metadata,roi_csv_json)

        out_files.insert(0,roi_csv)

    return {
        "roi_csv":roi_csv,
        "roi_output_dir":roi_output_dir,
        "output_dir":output_dir,
        "out_files":out_files
    }



class collate_csv_singleInputSpec(BaseInterfaceInputSpec):
    labels_dict = traits.Dict({},mandatory=False,desc='labels', usedefault=True)
    csv_list1 = traits.List(desc='list of files')
    csv_list2 = traits.List(desc='list of files')
    add_prefix = traits.Bool(False,desc="Create header prefix while joining tables",usedefault=True)

class collate_csv_singleOutputSpec(TraitedSpec):
    roi_csv = File(desc='CSV file of results')
    roi_output_dir = traits.String(desc='roi output dir')
    output_dir = traits.String(desc='output dir')
    out_files = traits.List(desc='list of files')
    
class collate_csv_single_pan(BaseInterface):
    input_spec = collate_csv_singleInputSpec
    output_spec = collate_csv_singleOutputSpec

    def _run_interface(self, runtime):

        # Call our python code here:
        outputs = collate_csv_single_proc(
            self.inputs.labels_dict,
            self.inputs.csv_list1,
            self.inputs.csv_list2,
            self.inputs.add_prefix
        )

        setattr(self, "_results", outputs)
        # And we are done
        return runtime

    def _list_outputs(self):
        return self._results


def create(labels_dict,name="collate_csv_single_node",csv_list1="",csv_list2="",add_prefix=False,LOGGER=IFLOGGER):
    # Create Node
    pan_node = Node(collate_csv_single_pan(), name=name)

    if LOGGER:
        LOGGER.info(f"Created Node {pan_node!r}")

    # Specify node inputs
    pan_node.inputs.labels_dict = labels_dict

    if not csv_list1 is None and not csv_list1 == "":
        pan_node.inputs.csv_list1 = csv_list1
 
    if not csv_list2 is None and not csv_list2 == "":
        pan_node.inputs.csv_list2 = csv_list2

    pan_node.inputs.add_prefix =  add_prefix

    return pan_node





from nipype.interfaces.base import BaseInterfaceInputSpec, BaseInterface, File, TraitedSpec, traits
from nipype import Node
from panpipelines.utils.util_functions import *
import os
import nibabel as nb
from bids import BIDSLayout
from nipype import logging as nlogging
import json

IFLOGGER=nlogging.getLogger('nipype.interface')

ASL_INPUT="-i"
ASL_OUTPUT="-o"
CALIBRATION="-c"
IAF="--iaf"
IBF="--ibf"
RPTS="--rpts"
CASL="--casl"
BOLUS="--bolus"
TIS="--tis"
TR="--tr"
CMETHOD="--cmethod"
FIXBOLUS="--fixbolus"
FIXBAT="--fixbat"
MC="--mc"
PVCORR="--pvcorr"
FSLANAT="--fslanat"
DEBUG="--debug"
REGIONANALYSIS="--region-analysis"
WHITEPAPER="--wp"

FMAP="--fmap"
FMAPMAG="--fmapmag"
FMAPMAGBRAIN="--fmapmagbrain"
ECHOSPACING="--echospacing"
PEDIR="--pedir"

IS_PRESENT="^^^"
IGNORE="###"

def splitASL(asl_input,command_base,work_dir,labels_dict,m0_index="0",m0_size="1",asl_index="1",asl_size="-1"):
    new_asl_input = newfile(outputdir=work_dir, assocfile=asl_input, suffix=f"desc-asl")
    params = f"{asl_input}"\
        f" {new_asl_input}" \
        f" {asl_index}" \
        f" {asl_size}" 

    command=f"{command_base} fslroi"\
        " "+params

    evaluated_command=substitute_labels(command, labels_dict)
    runCommand(evaluated_command,IFLOGGER)

    new_m0_input = newfile(outputdir=work_dir, assocfile=asl_input, suffix=f"desc-m0")
    params = f"{asl_input}"\
        f" {new_m0_input}" \
        f" {m0_index}" \
        f" {m0_size}" 

    command=f"{command_base} fslroi"\
        " "+params

    evaluated_command=substitute_labels(command, labels_dict)
    runCommand(evaluated_command,IFLOGGER)

    return new_asl_input, new_m0_input


def scaleASL(asl_input, command_base, scale_factor,work_dir, basil_dict,labels_dict):

    new_asl_input = newfile(outputdir=work_dir, assocfile=asl_input, suffix=f"scale-{scale_factor}")
    params = f"{asl_input}"\
        f" -mul {scale_factor}" \
        f" {new_asl_input}" \
        " -odt float"

    command=f"{command_base} fslmaths"\
        " "+params

    evaluated_command=substitute_labels(command, labels_dict)
    runCommand(evaluated_command,IFLOGGER)
    basil_dict = updateParams(basil_dict,ASL_INPUT,new_asl_input)
    
    return basil_dict

def process_sdcflows_fieldmap(fieldmap_dir,layout, asljson , asl_acq, basil_dict, labels_dict, command_base, work_dir,bids_dir,subject,session,fmap_mode="phasediff"):
    fmapjson = getGlob(os.path.join(fieldmap_dir,"*desc-preproc_fieldmap.json"))
    if not fmapjson:
        fmapjson = os.path.join(fieldmap_dir,f"sub-{subject}_ses-{session}_desc-preproc_fieldmap.json")
        fmapdict={}
        if fmap_mode == "phasediff":
            fmapdict["RawSources"] = getPhaseDiffSources(bids_dir,subject,session)
            fmapdict["Units"] = "Hz"
            export_labels(fmapdict,fmapjson)

    return process_fmriprep_fieldmap(fieldmap_dir,layout, asljson , asl_acq, basil_dict, labels_dict, command_base, work_dir,fmap_mode=fmap_mode)


def process_fmriprep_fieldmap(fmriprep_fieldmap_dir,layout, asljson , asl_acq, basil_dict, labels_dict, command_base, work_dir,fmap_mode="fmriprep"):
    IFLOGGER.info(f"Preparing fieldmap from {fmap_mode} for use in SDC.")
    fmap = getGlob(os.path.join(fmriprep_fieldmap_dir,"*desc-preproc_fieldmap.nii.gz"))
    fmaprads = newfile(work_dir,assocfile=fmap,intwix="desc-rads")
    fmapjson = getGlob(os.path.join(fmriprep_fieldmap_dir,"*desc-preproc_fieldmap.json"))

    IFLOGGER.info(f"Fieldmap located at {fmap}")
    fmapdict={}
    if fmapjson:
        with open(fmapjson,"r") as infile:
            fmapdict=json.load(infile)

    if "Units" in fmapdict.keys():
        if fmapdict["Units"] == "Hz":
            IFLOGGER.info(f"Convert {fmap} in Hz to {fmaprads} in rad/s")
            if "RawSources" in fmapdict.keys():
                sources = fmapdict["RawSources"]
                phase1 = [x for x in sources if "phase1" in x]
                echo1=None
                if phase1:
                    phase1_entities = layout.parse_file_entities(phase1[0])
                    phase1_md = layout.get(**phase1_entities)[0].get_metadata()
                    if phase1_md:
                        echo1=phase1_md["EchoTime"]
                phase2 = [x for x in sources if "phase2" in x]
                echo2=None
                if phase2:
                    phase2_entities = layout.parse_file_entities(phase2[0])
                    phase2_md = layout.get(**phase2_entities)[0].get_metadata()
                    if phase2_md:
                        echo2=phase2_md["EchoTime"]

                if echo1 and echo2:
                    echodiff = echo2 - echo1
                    mult = 2 * np.pi * np.abs(echodiff)

                    params = f"{fmap}"\
                        f" -mul {mult}" \
                        f" {fmaprads}" \
                        " -odt float"

                    command=f"{command_base} fslmaths"\
                        " "+params

                    evaluated_command=substitute_labels(command, labels_dict)
                    runCommand(evaluated_command,IFLOGGER)
                    basil_dict = updateParams(basil_dict,FMAP,fmaprads)
            
            elif fmapdict["Units"] == "rad/s":
                IFLOGGER.info(f"{fmap} ialready in rad/s. Rename to {fmaprads}.")
                fmaprads = fmap
                basil_dict = updateParams(basil_dict,FMAP,fmaprads)
            else:
                unkunits = fmapdict["Units"]
                IFLOGGER.info(f"Units {unkunits} not recognized.")
           
    
    if os.path.exists(fmaprads): 
        fmapmag = getGlob(os.path.join(fmriprep_fieldmap_dir,"*desc-magnitude_fieldmap.nii.gz"))
        fmapmag_brain = newfile(work_dir,assocfile=fmapmag,intwix="desc-brain")
        IFLOGGER.info(f"Create brain-extracted {fmapmag_brain} from {fmapmag}.")   

        params = f"{fmapmag}"\
            f" {fmapmag_brain}" \
            " -R"

        command=f"{command_base} bet"\
            " "+params

        evaluated_command=substitute_labels(command, labels_dict)
        runCommand(evaluated_command,IFLOGGER)

        basil_dict = updateParams(basil_dict,FMAPMAG,fmapmag)
        basil_dict = updateParams(basil_dict,FMAPMAGBRAIN,fmapmag_brain)

    if os.path.exists(fmapmag_brain):
        echospacing = None
        ECHOSPACING_DICT= getParams(labels_dict,"ASL_ECHOSPACING")
        if ECHOSPACING_DICT is not None and isinstance(ECHOSPACING_DICT, dict):
            if asl_acq in ECHOSPACING_DICT.keys():
                echospacing = ECHOSPACING_DICT[asl_acq]

        basil_dict = updateParams(basil_dict,ECHOSPACING,echospacing)
        if echospacing:
            IFLOGGER.info(f"Echospacing {echospacing} obtained from config file.")   
        else:
            IFLOGGER.error(f"Echospacing not defined. Fieldmap correction will not work.") 
            raise ValueError("<ASL_ECHOSPACING> not defined in config file for fieldmap processing.") 
        pedir_ijk = asljson["PhaseEncodingDirection"]
        IFLOGGER.info(f"PE Direction of ASL is {pedir_ijk}")   
        pedir_fsl = pedir_ijk.replace('j-','-y').replace('j','y').replace('i-','-x').replace('i','x').replace('k-','-z').replace('k','z')
        IFLOGGER.info(f"PE Direction of ASL converted to {pedir_fsl} for BASIL.")    
        basil_dict = updateParams(basil_dict,PEDIR,pedir_fsl)


    return basil_dict

# work in progress
def process_fsl_prepare_fieldmap(layout, asl_json,basil_dict,labels_dict, asljson, asl_acq, command_base, work_dir):

    participant_label = getParams(labels_dict,'PARTICIPANT_LABEL')

    phase1 = layout.get(subject=participant_label,suffix='phase1', extension='nii.gz')
    echo1=None
    if phase1:
        phase1_md = phase1[0].get_metadata()
        if phase1_md:
            echo1=phase1_md["EchoTime1"]

    phase2 = layout.get(subject=participant_label,suffix='phase2', extension='nii.gz')
    echo2=None
    if phase2:
        phase2_md = phase2[0].get_metadata()
        if phase2_md:
            echo2=phase2_md["EchoTime2"]

    mag1 = layout.get(subject=participant_label,suffix='magnitude1', extension='nii.gz')
    mag2 = layout.get(subject=participant_label,suffix='magnitude2', extension='nii.gz')
    return basil_dict

def basil_proc(labels_dict,bids_dir="",fslanat_dir=""):

    command_base, container = getContainer(labels_dict,nodename="basil",SPECIFIC="BASIL_CONTAINER",LOGGER=IFLOGGER)

    IFLOGGER.info("Checking the oxford_asl version:")
    command = f"{command_base} oxford_asl --version"
    evaluated_command=substitute_labels(command, labels_dict)
    runCommand(evaluated_command,IFLOGGER)

    if container:
        IFLOGGER.info("\nChecking the container version:")
        command = f"{command_base} --version"
        evaluated_command=substitute_labels(command, labels_dict)
        runCommand(evaluated_command,IFLOGGER)

    basil_dict={}
    basil_dict = updateParams(basil_dict,MC,IS_PRESENT)
    basil_dict = updateParams(basil_dict,PVCORR,IS_PRESENT)
    basil_dict = updateParams(basil_dict,DEBUG,IS_PRESENT)
    basil_dict = updateParams(basil_dict,REGIONANALYSIS,IS_PRESENT)
    basil_dict = updateParams(basil_dict,FIXBAT,IS_PRESENT)
    basil_dict = updateParams(basil_dict,WHITEPAPER,IS_PRESENT)
  
    cwd=os.getcwd()
    labels_dict = updateParams(labels_dict,"CWD",cwd)
    
    output_dir=os.path.join(cwd,"basiloutput")
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    basil_dict = updateParams(basil_dict,ASL_OUTPUT,output_dir)

    work_dir=os.path.join(output_dir,"basilwork_pan")
    if not os.path.exists(work_dir):
        os.makedirs(work_dir)

    participant_label = getParams(labels_dict,'PARTICIPANT_LABEL')
    participant_session = getParams(labels_dict,'PARTICIPANT_SESSION')
    
    layout = BIDSLayout(bids_dir)
    asl=layout.get(subject=participant_label,session=participant_session,suffix='asl', extension='nii.gz')

    if len(asl) > 0:
        asl_bidsfile=asl[0]
        asl_file=asl_bidsfile.path

        asljson=asl_bidsfile.get_metadata()
        asl_entities = asl_bidsfile.get_entities()
        if "acquisition" in asl_entities.keys():
            asl_acq = "acq-" + asl_entities["acquisition"]
        else:
            asl_acq = get_bidstag("acq",asl_bidsfile.filename)
        
        if not asl_acq:
            asl_acq = "default"

        asl_type="PCASL"
        if "ArterialSpinLabelingType" in asljson.keys():
            asl_type = asljson["ArterialSpinLabelingType"]

        m0_type="Separate"
        if "M0Type" in asljson.keys():
            m0_type = asljson["M0Type"]

        basil_dict = updateParams(basil_dict,IAF,"tc")
        ASLCONTEXT = getParams(labels_dict,"ASLCONTEXT")
        if ASLCONTEXT and isinstance(ASLCONTEXT,dict):
            if asl_acq in ASLCONTEXT.keys():
                if ASLCONTEXT[asl_acq] == "control:label":
                    basil_dict = updateParams(basil_dict,IAF,"ct")
                elif ASLCONTEXT[asl_acq] == "label:control":
                    basil_dict = updateParams(basil_dict,IAF,"tc")
                elif ASLCONTEXT[asl_acq] == "m0scan:control:label":
                    basil_dict = updateParams(basil_dict,IAF,"ct")
                    asl_file,m0_file = splitASL(asl_file,command_base,work_dir,labels_dict)

        asl_scaling = None
        ASL_SCALING_DICT = getParams(labels_dict,'ASL_SCALING')
        if ASL_SCALING_DICT  and isinstance(ASL_SCALING_DICT,dict):
            if asl_acq in ASL_SCALING_DICT.keys():
                asl_scaling = substitute_labels(ASL_SCALING_DICT[asl_acq],labels_dict)

        if asl_scaling:
            basil_dict = scaleASL(asl_file, command_base,asl_scaling, work_dir, basil_dict,labels_dict)
        else:
            basil_dict = updateParams(basil_dict,ASL_INPUT,asl_file)

        asl_img = nb.load(asl_file)
        rpts=int(asl_img.header["dim"][4]/2)
        basil_dict = updateParams(basil_dict,IBF,'rpt')
        basil_dict = updateParams(basil_dict,RPTS,str(rpts))

        fix_bolus=None
        if "BolusCutOffTechnique" in asljson.keys():
            fix_bolus = asljson["BolusCutOffTechnique"]

        if fix_bolus  is not None:
            basil_dict = updateParams(basil_dict,FIXBOLUS,IS_PRESENT)

        pld = None
        labelDuration=None
        if "PostLabelingDelay" in asljson.keys():
            pld = asljson["PostLabelingDelay"]
        if "LabelingDuration" in asljson.keys():
            labelDuration = asljson["LabelingDuration"]

        if labelDuration is not None:
            basil_dict = updateParams(basil_dict,BOLUS,str(labelDuration))

        if asl_type == "PCASL" or asl_type == "CASL":
            basil_dict = updateParams(basil_dict,CASL,IS_PRESENT)
            if labelDuration is not None and pld is not None:
                tis = pld + labelDuration
                basil_dict = updateParams(basil_dict,TIS,str(tis))
        else:
            # PLD in PASL json is actually the TIS
            if labelDuration is not None and pld is not None:
                tis = pld
                basil_dict = updateParams(basil_dict,TIS,str(tis))

        # process field map if it exists
        fieldmap_type=None
        FIELDMAP_TYPE_DICT = getParams(labels_dict,'FIELDMAP_TYPE')
        if FIELDMAP_TYPE_DICT  is not None and isinstance(FIELDMAP_TYPE_DICT,dict):
            if asl_acq in FIELDMAP_TYPE_DICT.keys():
                fieldmap_type = substitute_labels(FIELDMAP_TYPE_DICT[asl_acq],labels_dict)

        if fieldmap_type:
            # use preprocessed field mao from fmriprep
            if fieldmap_type == "fmriprep_preproc":
                fmriprep_fieldmap_dir = None
                FMRIPREP_FIELDMAP_DIR_DICT  = getParams(labels_dict,'FMRIPREP_FIELDMAP_DIR')
                if FMRIPREP_FIELDMAP_DIR_DICT is not None and isinstance(FMRIPREP_FIELDMAP_DIR_DICT ,dict):
                    if asl_acq in FMRIPREP_FIELDMAP_DIR_DICT.keys():
                        fmriprep_fieldmap_dir = substitute_labels(FMRIPREP_FIELDMAP_DIR_DICT[asl_acq],labels_dict)

                if fmriprep_fieldmap_dir and os.path.exists(fmriprep_fieldmap_dir):
                    basil_dict = process_fmriprep_fieldmap(fmriprep_fieldmap_dir,layout, asljson , asl_acq, basil_dict, labels_dict, command_base, work_dir)
                else:
                    IFLOGGER.warn(f"Attempting to create fieldmap using {fieldmap_type} but sources not found. Exiting")
                    sys.exit(1)
            elif fieldmap_type == "sdcflows_preproc":
                fieldmap_dir = None
                SDCFLOWS_FIELDMAP_DIR_DICT  = getParams(labels_dict,'SDCFLOWS_FIELDMAP_DIR')
                if SDCFLOWS_FIELDMAP_DIR_DICT is not None and isinstance(SDCFLOWS_FIELDMAP_DIR_DICT ,dict):
                    if asl_acq in SDCFLOWS_FIELDMAP_DIR_DICT.keys():
                        fieldmap_dir = substitute_labels(SDCFLOWS_FIELDMAP_DIR_DICT[asl_acq],labels_dict)

                SDCFLOWS_FMAP_MODE = getParams(labels_dict,"SDCFLOWS_FIELDMAP_MODE") 
                if SDCFLOWS_FMAP_MODE:
                    sdcflows_fmap_mode = SDCFLOWS_FMAP_MODE
                else:
                    sdcflows_fmap_mode="phasediff"
                    labels_dict = updateParams(labels_dict,"SDCFLOWS_FIELDMAP_MODE",sdcflows_fmap_mode)

                if fieldmap_dir and os.path.exists(fieldmap_dir):
                    basil_dict = process_sdcflows_fieldmap(fieldmap_dir,layout, asljson , asl_acq, basil_dict, labels_dict, command_base, work_dir,bids_dir=bids_dir,subject=participant_label,session=participant_session, fmap_mode=sdcflows_fmap_mode)
                else:
                    IFLOGGER.warn(f"Attempting to create fieldmap using {fieldmap_type} but sources not found. Exiting")
                    sys.exit(1)
            elif fieldmap_type == "fsl_prepare_fieldmap":
                IFLOGGER.warn(f"{fieldmap_type} not yet implemented. Exiting")
                sys.exit(1)              
                #basil_dict = process_fsl_prepare_fieldmap(layout, asljson,basil_dict,labels_dict, asljson, asl_acq, command_base, work_dir)

        fslanat_dir=os.path.abspath(fslanat_dir)
        basil_dict = updateParams(basil_dict,FSLANAT,fslanat_dir)


        CMETHOD_OPTS = getParams(labels_dict,"CMETHOD_OPTS")
        if CMETHOD_OPTS is not None and isinstance(CMETHOD_OPTS,dict):
            if asl_acq in CMETHOD_OPTS.keys():
                basil_dict = updateParams(basil_dict,CMETHOD,CMETHOD_OPTS[asl_acq])

        if m0_type == "Separate":
            m0_entities = asl_entities.copy()
            m0_entities["suffix"]="m0scan"
            m0  = layout.get(return_type='file', invalid_filters='allow', **m0_entities)
            if len(m0) > 0:
                m0_file=m0[0]
                basil_dict = updateParams(basil_dict,CALIBRATION,m0_file)
                m0_md = layout.get(**m0_entities)[0].get_metadata()
                if "RepetitionTime" in m0_md.keys():
                    basil_dict = updateParams(basil_dict,TR,str(m0_md["RepetitionTime"]))
                else:
                    IFLOGGER.warn("RepetitionTime not found for m0. Default of 5s will be used by BASIL.")
        elif m0_type == "Included" and m0_file:
            IFLOGGER.info(f"M0 defined as included and defined at {m0_file}.")
            basil_dict = updateParams(basil_dict,CALIBRATION,m0_file)
            if "RepetitionTime" in asljson.keys():
                basil_dict = updateParams(basil_dict,TR,str(asljson["RepetitionTime"]))
            else:
                IFLOGGER.warn("RepetitionTime not found for m0. Default of 5s will be used by BASIL.")

        
        else:
            IFLOGGER.info(f"M0 Type {m0_type} not recognized. Expecting Separate or Included. Check spelling and case.")


        # Additional params
        BASIL_OVERRIDE_PARAMS = getParams(labels_dict,"BASIL_OVERRIDE_PARAMS")
        if BASIL_OVERRIDE_PARAMS and isinstance(BASIL_OVERRIDE_PARAMS,dict):
            add_labels(BASIL_OVERRIDE_PARAMS,basil_dict)        

        params = ""
        for basil_tag, basil_value in basil_dict.items():
            if "--" in basil_tag and "---" not in basil_tag:
                if basil_value == IS_PRESENT:
                    params=params + " " + basil_tag
                elif basil_value == IGNORE:
                    IFLOGGER.info(f"Parameter {basil_tag} is being skipped. This has been explicitly required in configuration.")
                else:
                    params = params + " " + basil_tag+"="+basil_value

            elif "-" in basil_tag and "--" not in basil_tag:
                params = params + " " + basil_tag + " " + basil_value

            else:
                print(f"Basil tag {basil_tag} not valid.")


        command=f"{command_base} oxford_asl"\
            " "+params

        evaluated_command=substitute_labels(command, labels_dict)
        runCommand(evaluated_command,IFLOGGER)
    else:
        IFLOGGER.warn(f"ASL acquisition not found for subject {participant_label} and session {participant_session}")


    cbf_native = getGlob(os.path.join(output_dir,"native_space","perfusion_calib.nii.gz"))
    cbf_t1 = getGlob(os.path.join(output_dir,"struct_space","perfusion_calib.nii.gz"))
    cbf_mni6 = getGlob(os.path.join(output_dir,"std_space","perfusion_calib.nii.gz"))
    out_files=[]
    out_files.insert(0,cbf_native)
    out_files.insert(1,cbf_t1)
    out_files.insert(2,cbf_mni6)


    return {
        "cbf_native":cbf_native,
        "cbf_t1":cbf_t1,
        "cbf_mni6":cbf_mni6,
        "output_dir":output_dir,
        "out_files":out_files
    }



class basilInputSpec(BaseInterfaceInputSpec):
    labels_dict = traits.Dict({},mandatory=False,desc='labels', usedefault=True)
    bids_dir = traits.String("",desc="BIDS Directory", usedefault=True)
    fslanat_dir = traits.String("",desc="FSLANAT Directory", usedefault=True)

class basilOutputSpec(TraitedSpec):
    cbf_native = File(desc='cbf_native')
    cbf_t1 = File(desc='cbf_t1')
    cbf_mni6 = File(desc='cbf_mni6')    
    output_dir = traits.String(desc="output directory of basil output")
    out_files = traits.List(desc='list of files')
    
class basil_pan(BaseInterface):
    input_spec = basilInputSpec
    output_spec = basilOutputSpec

    def _run_interface(self, runtime):

        # Call our python code here:
        outputs = basil_proc(
            self.inputs.labels_dict,
            self.inputs.bids_dir,
            self.inputs.fslanat_dir
        )

        setattr(self, "_results", outputs)
        # And we are done
        return runtime

    def _list_outputs(self):
        return self._results


def create(labels_dict,name="basil_node",bids_dir="",fslanat_dir="",LOGGER=IFLOGGER):
    # Create Node
    pan_node = Node(basil_pan(), name=name)

    if LOGGER:
        LOGGER.info(f"Created Node {pan_node!r}")

    # Specify node inputs
    pan_node.inputs.labels_dict = labels_dict

    if bids_dir is None or bids_dir == "":
        bids_dir = substitute_labels("<BIDS_DIR>", labels_dict)

    pan_node.inputs.bids_dir =  bids_dir
    pan_node.inputs.fslanat_dir =  fslanat_dir

    return pan_node





__all__ = [
    "qsiprep_workflow",
    "tensor_workflow",
    "noddi_workflow",
    "freesurfer_workflow",
    "fmriprep_workflow",
    "basil_workflow",
    "aslprep_workflow",
    "dummy_workflow",
    "volmeasures_workflow",
    "roiextract_workflow",
    "textmeasures_workflow",
    "collatecsv_workflow",
    "collatecsvgroup_workflow",
    "lst_workflow",
    "preproc_workflow"
]


from panpipelines.utils.util_functions import *
from panpipelines.scripts.panscript import *


class pancontainer_panscript(panscript):

    def __init__(self,labels_dict,name='pancontainer_panscript',params="",command="",container_img="PAN_CONTAINER",execution={}):
        super().__init__(labels_dict,name=name,params=params,command=command)

        self.params = params

        if not container_img:
            container_img = "PAN_CONTAINER"

        command_base, container = getContainer(labels_dict,nodename="pancontainer_panscript",SPECIFIC=container_img,LOGGER=IFLOGGER)

        if container:
            IFLOGGER.info("\nChecking the container version:")
            version_command = f"{command_base} --version"
            evaluated_command=substitute_labels(version_command, labels_dict)
            runCommand(evaluated_command,IFLOGGER)

        
        self.command = command_base + " " + command










from nipype import Workflow

import panpipelines.nodes.basil as basil
import panpipelines.nodes.fslanat as fslanat
from panpipelines.utils.util_functions import *
from bids import BIDSLayout
from sdcflows import fieldmaps as sfm
from nipype import Workflow, Node
from nipype.interfaces.io import DataSink
import panpipelines.scripts.pancontainer_panscript as pancontainer_script
import panpipelines.scripts.sdcflows_fieldmap as sdcflows_fieldmap

def create(name, wf_base_dir,labels_dict,createGraph=True,execution={}, LOGGER=None):
    # Create workflow
    pan_workflow = Workflow(name=name, base_dir=wf_base_dir)

    if LOGGER:
        LOGGER.info(f"Created Workflow {name} with base directory {wf_base_dir}")

    if len(execution.keys()) > 0:
        pan_workflow.config = process_dict(pan_workflow.config,execution)

    # Specify node inputs
    fslanat_manual=getParams(labels_dict,"FSLANAT_MANUAL")
    if fslanat_manual and fslanat_manual == "Y":
        fslanat_dir=getParams(labels_dict,"FSLANAT_DIR")
        basil_node = basil.create(labels_dict,fslanat_dir=fslanat_dir,LOGGER=LOGGER)
        pan_workflow.add_nodes([basil_node])
    else:
        fslanat_node = fslanat.create(labels_dict,LOGGER=LOGGER)
        basil_node = basil.create(labels_dict,LOGGER=LOGGER)
        pan_workflow.connect(fslanat_node,'fslanat_dir',basil_node,'fslanat_dir')

    FIELDMAP_TYPE = getParams(labels_dict,"FIELDMAP_TYPE")
    SDCFLOWS_FMAP_DIR = getParams(labels_dict,"SDCFLOWS_FIELDMAP_DIR")
    if FIELDMAP_TYPE and SDCFLOWS_FMAP_DIR:
        bids_dir = getParams(labels_dict,"BIDS_DIR")
        participant_label = getParams(labels_dict,'PARTICIPANT_LABEL')
        participant_session = getParams(labels_dict,'PARTICIPANT_SESSION')
        acq_label = getAcquisition(bids_dir,participant_label,participant_session=participant_session)
        if isinstance(FIELDMAP_TYPE,dict):
            if acq_label in FIELDMAP_TYPE.keys():
                FIELDMAP_TYPE = FIELDMAP_TYPE[acq_label]   

        if FIELDMAP_TYPE == "sdcflows_preproc":
            if isinstance(SDCFLOWS_FMAP_DIR,dict):
                if acq_label in SDCFLOWS_FMAP_DIR.keys():
                    SDCFLOWS_FMAP_DIR = SDCFLOWS_FMAP_DIR[acq_label]
                else:
                    SDCFLOWS_FMAP_DIR[acq_label]=substitute_labels("<WORKFLOW_DIR>/sdcflows/fmap",labels_dict) 
                    labels_dict = updateParams(labels_dict,"SDCFLOWS_FIELDMAP_DIR",SDCFLOWS_FMAP_DIR)
                    SDCFLOWS_FMAP_DIR = SDCFLOWS_FMAP_DIR[acq_label]

            # we have to mnanually substitute items from DICTs until we also check these in update_labels
            SDCFLOWS_FMAP_DIR=substitute_labels(SDCFLOWS_FMAP_DIR,labels_dict)
    
            SDCFLOWS_FMAP_MODE = getParams(labels_dict,"SDCFLOWS_FIELDMAP_MODE") 
            if SDCFLOWS_FMAP_MODE:
                sdcflows_fmap_mode = SDCFLOWS_FMAP_MODE
            else:
                sdcflows_fmap_mode="phasediff"
                labels_dict = updateParams(labels_dict,"SDCFLOWS_FIELDMAP_MODE",sdcflows_fmap_mode)
            
            sdcflows_workdir = os.path.dirname(SDCFLOWS_FMAP_DIR)
            if not os.path.exists(sdcflows_workdir):
                os.makedirs(sdcflows_workdir)

            sdcflows_fmap_parentdir = os.path.dirname(SDCFLOWS_FMAP_DIR)
            if not os.path.exists(sdcflows_fmap_parentdir):
                os.makedirs(sdcflows_fmap_parentdir)

            sources=[]
            if sdcflows_fmap_mode == "phasediff":               
                sources = getPhaseDiffSources(bids_dir,participant_label,participant_session)
                sources_String = " ".join(sources)
                params = f"--fmap_sources {sources_String}" \
                    f" --subject {participant_label}" \
                    f" --session {participant_session}" \
                    f" --fieldmap_dir {SDCFLOWS_FMAP_DIR }" \
                    f" --workdir {sdcflows_workdir}"

            if sources:
                SDCFLOWS_CONTAINER_TO_USE = getParams(labels_dict,"SDCFLOWS_CONTAINER_TO_USE")
                if SDCFLOWS_CONTAINER_TO_USE:
                    panscript = pancontainer_script.pancontainer_panscript(labels_dict,params=params,command=f"python {sdcflows_fieldmap.__file__}",container_img=SDCFLOWS_CONTAINER_TO_USE)
                else: 
                    panscript = pancontainer_script.pancontainer_panscript(labels_dict,params=params,command=f"python {sdcflows_fieldmap.__file__}")
                panscript.run()
            else:
                if LOGGER:
                    LOGGER.warn("Attempting to create fieldmap but sources not found. Exiting")
                sys.exit(1)

            
    if createGraph:
         pan_workflow.write_graph(graph2use='flat')

    return pan_workflow



from nipype.interfaces.base import BaseInterfaceInputSpec, BaseInterface, File, TraitedSpec, traits
from nipype import Node
from panpipelines.utils.util_functions import *
import os
import glob
import shlex
import subprocess
from nipype import logging as nlogging

IFLOGGER=nlogging.getLogger('nipype.interface')

IS_PRESENT="^^^"
IGNORE="###"

def noddi_proc(labels_dict,input_dir):

    cwd=os.getcwd()
    labels_dict = updateParams(labels_dict,"CWD",cwd)

    TEMPLATEFLOW_HOME=getParams(labels_dict,"TEMPLATEFLOW_HOME")
    os.environ["TEMPLATEFLOW_HOME"]=TEMPLATEFLOW_HOME
    os.environ["SINGULARITYENV_TEMPLATEFLOW_HOME"]=TEMPLATEFLOW_HOME
    
    command_base, container = getContainer(labels_dict,nodename="qsiprep", SPECIFIC="QSIPREP_CONTAINER",LOGGER=IFLOGGER)
    IFLOGGER.info("Checking the qsiprep version:")
    command = f"{command_base} --version"
    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    input_dir=substitute_labels(input_dir,labels_dict)

    # set up dwi to process just the specific dwi session
    participant_label = getParams(labels_dict,'PARTICIPANT_LABEL')
    participant_session = getParams(labels_dict,'PARTICIPANT_SESSION')

    bids_filter_dict={}
    bids_filter_dict["dwi"] = {}
    bids_filter_dict["dwi"]["session"] =  participant_session
    bids_filter_file = os.path.join(cwd,f"{participant_label}_{participant_session}_bids_filter_file.json")
    export_labels(bids_filter_dict,bids_filter_file)
    IFLOGGER.info(f"Specifying session filter: exporting {bids_filter_dict} to {bids_filter_file}")

    qsirecon_dict={}
    qsirecon_dict = updateParams(qsirecon_dict,"--participant_label","<PARTICIPANT_LABEL>")
    qsirecon_dict = updateParams(qsirecon_dict,"--recon_input",input_dir)
    qsirecon_dict = updateParams(qsirecon_dict,"--recon_spec","<RECON_TYPE>")
    qsirecon_dict = updateParams(qsirecon_dict,"--recon-only",IS_PRESENT)
    qsirecon_dict = updateParams(qsirecon_dict,"--bids-filter-file",bids_filter_file)
    qsirecon_dict = updateParams(qsirecon_dict,"--mem_mb","<BIDSAPP_MEMORY>")
    qsirecon_dict = updateParams(qsirecon_dict,"--nthreads","<BIDSAPP_THREADS>")
    qsirecon_dict = updateParams(qsirecon_dict,"--fs-license-file","<FSLICENSE>")
    qsirecon_dict = updateParams(qsirecon_dict,"--skip-odf-report",IS_PRESENT)
    qsirecon_dict = updateParams(qsirecon_dict,"--output-resolution","<OUTPUT_RES>")
    qsirecon_dict = updateParams(qsirecon_dict,"-w","<CWD>/noddi_work")

    # Additional params
    QSIRECON_OVERRIDE_PARAMS = getParams(labels_dict,"QSIRECON_OVERRIDE_PARAMS")
    if QSIRECON_OVERRIDE_PARAMS and isinstance(QSIRECON_OVERRIDE_PARAMS,dict):
        add_labels(QSIRECON_OVERRIDE_PARAMS,qsirecon_dict)        

    params = ""
    for qsirecon_tag, qsirecon_value in qsirecon_dict.items():
        if "--" in qsirecon_tag and "---" not in qsirecon_tag:
            if qsirecon_value == IS_PRESENT:
                params=params + " " + qsirecon_tag
            elif qsirecon_value == IGNORE:
                IFLOGGER.info(f"Parameter {qsirecon_tag} is being skipped. This has been explicitly required in configuration.")
            else:
                # we dont need = sign for qsirecon just for basi;
                params = params + " " + qsirecon_tag + " " + qsirecon_value

        elif "-" in qsirecon_tag and "--" not in qsirecon_tag:
            params = params + " " + qsirecon_tag + " " + qsirecon_value

        else:
            print(f"qsirecon tag {qsirecon_tag} not valid.") 

    command=f"{command_base}"\
            " <BIDS_DIR>"\
            " <CWD>"\
            " participant"\
            " "+params

    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    participant_label = getParams(labels_dict,'PARTICIPANT_LABEL')
    icvf = getGlob(os.path.join(cwd,'qsirecon','sub-{}'.format(participant_label),'ses-*','dwi','*ICVF_NODDI.nii.gz'))
    isovf = getGlob(os.path.join(cwd,'qsirecon','sub-{}'.format(participant_label),'ses-*','dwi','*ISOVF_NODDI.nii.gz'))
    od = getGlob(os.path.join(cwd,'qsirecon','sub-{}'.format(participant_label),'ses-*','dwi','*OD_NODDI.nii.gz'))
    directions =  getGlob(os.path.join(cwd,'qsirecon','sub-{}'.format(participant_label),'ses-*','dwi','*directions_NODDI.nii.gz'))
    output_dir = cwd

    
    out_files=[]
    out_files.insert(0,icvf)
    out_files.insert(1,isovf)
    out_files.insert(2,od)
    out_files.insert(3,directions)


    return {
        "icvf":icvf,
        "isovf":isovf,
        "od":od,
        "directions":directions,
        "output_dir":output_dir,
        "out_files":out_files
    }



class noddiInputSpec(BaseInterfaceInputSpec):
    labels_dict = traits.Dict({},mandatory=False,desc='labels', usedefault=True)
    input_dir = traits.String("",desc="BIDS Directory", usedefault=True)

class noddiOutputSpec(TraitedSpec):
    icvf = File(desc='ICVF')
    isovf = File(desc='ISOVF')
    od = File(desc='OD')
    directions = File(desc='Directions')
    output_dir = traits.String(desc="NODDI output directory")
    out_files = traits.List(desc='list of files')
    
class noddi_pan(BaseInterface):
    input_spec = noddiInputSpec
    output_spec = noddiOutputSpec

    def _run_interface(self, runtime):

        # Call our python code here:
        outputs = noddi_proc(
            self.inputs.labels_dict,
            self.inputs.input_dir
        )

        setattr(self, "_results", outputs)
        # And we are done
        return runtime

    def _list_outputs(self):
        return self._results


def create(labels_dict,name="noddi_node",input_dir="", LOGGER=IFLOGGER):
    # Create Node
    pan_node = Node(noddi_pan(), name=name)

    if LOGGER:
        LOGGER.info(f"Created Node {pan_node!r}") 
           
    # Specify node inputs

    labels_dict = updateParams(labels_dict,"RECON_TYPE","amico_noddi")
    pan_node.inputs.labels_dict = labels_dict

    if input_dir is None or input_dir == "":
        input_dir = substitute_labels("<QSIPREP_OUTPUT>", labels_dict)
        
    pan_node.inputs.input_dir =  input_dir

    return pan_node





from nipype.interfaces.base import BaseInterfaceInputSpec, BaseInterface, File, TraitedSpec, traits
from nipype import Node
from panpipelines.utils.util_functions import *
import os
import glob
import shlex
import subprocess
from nipype import logging as nlogging

IFLOGGER=nlogging.getLogger('nipype.interface')

IS_PRESENT="^^^"
IGNORE="###"

def aslprep_proc(labels_dict,bids_dir=""):

    TEMPLATEFLOW_HOME=getParams(labels_dict,"TEMPLATEFLOW_HOME")
    os.environ["TEMPLATEFLOW_HOME"]=TEMPLATEFLOW_HOME
    os.environ["SINGULARITYENV_TEMPLATEFLOW_HOME"]=TEMPLATEFLOW_HOME
    
    command_base, container = getContainer(labels_dict,nodename="aslprep",SPECIFIC="ASLPREP_CONTAINER",LOGGER=IFLOGGER)
    IFLOGGER.info("Checking the aslprep version:")
    command = f"{command_base} --version"
    evaluated_command=substitute_labels(command,labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    cwd=os.getcwd()
    labels_dict = updateParams(labels_dict,"CWD",cwd)
    
    participant_label = getParams(labels_dict,'PARTICIPANT_LABEL')
    TEMPLATEFLOW_HOME=getParams(labels_dict,"TEMPLATEFLOW_HOME")
    os.environ["TEMPLATEFLOW_HOME"]=TEMPLATEFLOW_HOME
    os.environ["SINGULARITYENV_TEMPLATEFLOW_HOME"]=TEMPLATEFLOW_HOME

    aslprep_dict={}
    aslprep_dict = updateParams(aslprep_dict,"--participant_label","<PARTICIPANT_LABEL>")
    aslprep_dict = updateParams(aslprep_dict,"--low-mem",IS_PRESENT)
    aslprep_dict = updateParams(aslprep_dict,"--skip-bids-validation",IS_PRESENT)
    aslprep_dict = updateParams(aslprep_dict,"--stop-on-first-crash",IS_PRESENT)
    aslprep_dict = updateParams(aslprep_dict,"--mem_mb","<BIDSAPP_MEMORY>")
    aslprep_dict = updateParams(aslprep_dict,"--nthreads","<BIDSAPP_THREADS>")
    aslprep_dict = updateParams(aslprep_dict,"--fs-license-file","<FSLICENSE>")
    aslprep_dict = updateParams(aslprep_dict,"--ignore fieldmaps",IS_PRESENT)
    aslprep_dict = updateParams(aslprep_dict,"-w","<CWD>/aslprep_work")

    # Additional params
    ASLPREP_OVERRIDE_PARAMS = getParams(labels_dict,"ASLPREP_OVERRIDE_PARAMS")
    if ASLPREP_OVERRIDE_PARAMS and isinstance(ASLPREP_OVERRIDE_PARAMS,dict):
        add_labels(ASLPREP_OVERRIDE_PARAMS,aslprep_dict)        

    params = ""
    for aslprep_tag, aslprep_value in aslprep_dict.items():
        if "--" in aslprep_tag and "---" not in aslprep_tag:
            if aslprep_value == IS_PRESENT:
                params=params + " " + aslprep_tag
            elif aslprep_value == IGNORE:
                IFLOGGER.info(f"Parameter {aslprep_tag} is being skipped. This has been explicitly required in configuration.")
            else:
                # we dont need = sign for fmriprep just for basi;
                params = params + " " + aslprep_tag + " " + aslprep_value

        elif "-" in aslprep_tag and "--" not in aslprep_tag:
            params = params + " " + aslprep_tag + " " + aslprep_value

        else:
            print(f"aslprep tag {aslprep_tag} not valid.") 

    command = f"{command_base}"\
        " <BIDS_DIR>"\
        " <CWD>"\
        " participant"\
        " " + params
        
    evaluated_command=substitute_labels(command,labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    cbf_preprocess = getGlob(os.path.join(cwd,'aslprep','sub-{}'.format(participant_label),'ses-*','perf','*cbf.nii.gz'))
    mat_t12perf = getGlob(os.path.join(cwd,'aslprep','sub-{}'.format(participant_label),'ses-*','perf','*from-T1w*mode-image_xfm.txt'))
    mat_perf2t1 = getGlob(os.path.join(cwd,'aslprep','sub-{}'.format(participant_label),'ses-*','perf','*from-scanner*mode-image_xfm.txt'))
    mat_t12mni = getGlob(os.path.join(cwd,'aslprep','sub-{}'.format(participant_label),'ses-*','anat','*from-T1w*mode-image_xfm.h5'))
    mat_mni2t1 = getGlob(os.path.join(cwd,'aslprep','sub-{}'.format(participant_label),'ses-*','anat','*from-MNI*mode-image_xfm.h5'))
    anatrefs =  glob.glob(os.path.join(cwd,'aslprep','sub-{}'.format(participant_label),'ses-*','anat','*{}*_desc-preproc_T1w.nii.gz'.format(participant_label)))
    t1ref = getFirstFromList([ s for s in anatrefs if "MNI" not in s])
    mniref = getFirstFromList([ s for s in anatrefs if "MNI" in s])
    output_dir = cwd

    
    out_files=[]
    out_files.insert(0,cbf_preprocess)
    out_files.insert(1,mat_t12perf)
    out_files.insert(2,mat_perf2t1)
    out_files.insert(3,mat_t12mni)
    out_files.insert(4,mat_mni2t1)
    out_files.insert(5,mniref)
    out_files.insert(6,t1ref)


    return {
        "cbf_preprocess":cbf_preprocess,
        "mat_t12perf":mat_t12perf,
        "mat_perf2t1":mat_perf2t1,
        "mat_t12mni":mat_t12mni,
        "mat_mni2t1":mat_mni2t1,
        "mniref":mniref,
        "t1ref":t1ref,
        "output_dir":output_dir,
        "out_files":out_files
    }



class aslprepInputSpec(BaseInterfaceInputSpec):
    labels_dict = traits.Dict({},mandatory=False,desc='labels', usedefault=True)
    bids_dir = traits.String("",desc="BIDS Directory", usedefault=True)

class aslprepOutputSpec(TraitedSpec):
    cbf_preprocess = File(desc='Preprocessed CBF')
    mat_t12perf = File(desc='T1 to CBF/perfusion transform')
    mat_perf2t1 = File(desc='CBF/perfusion to T1 transform')
    mat_t12mni = File(desc='T1 to MNI transform')
    mat_mni2t1 = File(desc='MNI to T1 transform')
    mniref = File(desc='MNI reference')
    t1ref = File(desc='T1 reference')
    output_dir = traits.String(desc="ASLPREP output directory")
    out_files = traits.List(desc='list of files')
    
class aslprep_pan(BaseInterface):
    input_spec = aslprepInputSpec
    output_spec = aslprepOutputSpec

    def _run_interface(self, runtime):

        # Call our python code here:
        outputs = aslprep_proc(
            self.inputs.labels_dict,
            self.inputs.bids_dir
        )

        setattr(self, "_results", outputs)
        # And we are done
        return runtime

    def _list_outputs(self):
        return self._results


def create(labels_dict,name='aslprep_node',bids_dir="", LOGGER=IFLOGGER):
    # Create Node
    pan_node = Node(aslprep_pan(), name=name)

    if LOGGER:
        LOGGER.info(f"Created Node {pan_node!r}")

    # Specify node inputs
    pan_node.inputs.labels_dict = labels_dict
    
    if bids_dir is None or bids_dir == "":
        bids_dir = substitute_labels("<BIDS_DIR>", labels_dict)

    pan_node.inputs.bids_dir =  bids_dir

    return pan_node





from nipype.interfaces.base import BaseInterfaceInputSpec, BaseInterface, File, TraitedSpec, traits
from nipype import Node
from panpipelines.utils.util_functions import *
from panpipelines.utils.transformer import *
import os
import glob
import nibabel as nb
from nipype import logging as nlogging

IFLOGGER=nlogging.getLogger('nipype.interface')

def atlascreate_proc(labels_dict,roi_list,roilabels_list):

    cwd=os.getcwd()
    labels_dict = updateParams(labels_dict,"CWD",cwd)
    output_dir = cwd
 
    participant_label = getParams(labels_dict,'PARTICIPANT_LABEL')

    atlas_name = getParams(labels_dict,'ATLAS_NAME')
    atlas_workdir = os.path.join(cwd,'{}_workdir'.format(atlas_name))
    if not os.path.isdir(atlas_workdir):
        os.makedirs(atlas_workdir)

    atlas_file = newfile(cwd, atlas_name, prefix=f"sub-{participant_label}", extension="nii.gz")
    IFLOGGER.info(f"Creating new atlas {atlas_file}")

    special_atlas_type=""
    # scan through the roi list and find out if we have a special atlas type
    if roi_list[0] == "hcpmmp1aseg":
        special_atlas_type="hcpmmp1aseg"


    atlas_type = "3D"
    atlas_type = getParams(labels_dict,'NEWATLAS_TYPE')

    atlas_index_mode = None
    if getParams(labels_dict,'NEWATLAS_INDEX_MODE'):
        atlas_index_mode = getParams(labels_dict,'NEWATLAS_INDEX_MODE')
    elif getParams(labels_dict,'ATLAS_INDEX_MODE'):
        atlas_index_mode = getParams(labels_dict,'ATLAS_INDEX_MODE')

    if special_atlas_type == "hcpmmp1aseg":
        roilabels_list=create_3d_hcpmmp1_aseg(atlas_file,roi_list,labels_dict)
        if not atlas_index_mode:
            atlas_index_mode = "hcpmmp1aseg_tsv"
        roi_list = [atlas_file]
    elif atlas_type == "3D":
        create_3d_atlas_from_rois(atlas_file, roi_list,labels_dict)
    elif atlas_type == "3D_contig":
        create_3d_atlas_from_rois(atlas_file, roi_list,labels_dict,explode3d=False)
    elif atlas_type =="4D":
        create_4d_atlas_from_rois(atlas_file, roi_list,labels_dict)
    else:
        create_3d_atlas_from_rois(atlas_file, roi_list,labels_dict)

    if not atlas_index_mode:
        atlas_index_mode = "tsv"

    if "tsv" in atlas_index_mode:
        atlas_index = newfile(cwd, atlas_name, prefix=f"sub-{participant_label}", suffix="dseg",extension="tsv")
    else:
        atlas_index = newfile(cwd, atlas_name, prefix=f"sub-{participant_label}",suffix="dseg", extension="txt")
    atlas_index_json = newfile(cwd,atlas_index,extension="json")
    IFLOGGER.info(f"Creating new atlas index {atlas_index}")
    atlas_index_out=""
    atlas_dict={}
    atlas_dict["Generator"]=roilabels_list
    with open(atlas_index,"w") as outfile:
        for roi_num in range(len(roilabels_list)):
            roiname=roilabels_list[roi_num]
            if roiname.split(":")[0] == "get_freesurfer_atlas_index":
                lutfile = substitute_labels(roiname.split(":")[1],labels_dict)

                atlas_dict,atlas_index_out=get_freesurferatlas_index_mode(roi_list[roi_num],lutfile,None,atlas_index_mode)
                atlas_dict["Generator"]=roilabels_list
                break
            else:
                if roi_num== 0 and "tsv" in atlas_index_mode:
                    atlas_index_out = atlas_index_out + "index\tlabel\n"

                if "tsv" in atlas_index_mode:
                    atlas_index_out = atlas_index_out + f"{roi_num + 1}\t{roiname}\n"
                else:
                    atlas_index_out = atlas_index_out  + roiname + "\n"
        outfile.write(atlas_index_out)
        export_labels(atlas_dict,atlas_index_json)

    out_files=[]
    out_files.insert(0,atlas_file)
    out_files.insert(0,atlas_index)


    return {
        "atlas_file":atlas_file,
        "atlas_index":atlas_index,
        "output_dir":output_dir,
        "out_files":out_files
    }



class atlascreateInputSpec(BaseInterfaceInputSpec):
    labels_dict = traits.Dict({},mandatory=False,desc='labels', usedefault=True)
    roi_list = traits.List(desc='list of roi files')
    roilabels_list = traits.List(desc='list of labels')

class atlascreateOutputSpec(TraitedSpec):
    atlas_file = File(desc='new atlas file')
    atlas_index = File(desc='new atlas index file')
    output_dir = traits.String(desc="new atlas output directory")
    out_files = traits.List(desc='list of files')
    
class atlascreate_pan(BaseInterface):
    input_spec = atlascreateInputSpec
    output_spec = atlascreateOutputSpec

    def _run_interface(self, runtime):

        # Call our python code here:
        outputs = atlascreate_proc(
            self.inputs.labels_dict,
            self.inputs.roi_list,
            self.inputs.roilabels_list
        )

        setattr(self, "_results", outputs)
        # And we are done
        return runtime

    def _list_outputs(self):
        return self._results


def create(labels_dict,name="atlascreate_node",roi_list="",roilabels_list="", LOGGER=IFLOGGER):
    # Create Node
    pan_node = Node(atlascreate_pan(), name=name)

    if LOGGER:
        LOGGER.info(f"Created Node {pan_node!r}")

    # Specify node inputs
    pan_node.inputs.labels_dict = labels_dict
    pan_node.inputs.roi_list  =  roi_list
    pan_node.inputs.roilabels_list = roilabels_list

    return pan_node





from nipype.interfaces.base import BaseInterfaceInputSpec, BaseInterface, File, TraitedSpec, traits
from nipype import Node
from panpipelines.utils.util_functions import *
import os
import glob
from bids import BIDSLayout
import shlex
import subprocess
from nipype import logging as nlogging

IFLOGGER=nlogging.getLogger('nipype.interface')

def freesurfer_proc(labels_dict,bids_dir=""):

    participant_label = getParams(labels_dict,'PARTICIPANT_LABEL')
    session_label = getParams(labels_dict,'PARTICIPANT_SESSION')
    layout = BIDSLayout(bids_dir)

    subject="sub-"+participant_label
    cwd=os.getcwd()
    labels_dict = updateParams(labels_dict,"CWD",cwd)
    
    subjects_dir = os.path.join(cwd,'subjects_dir')
    if not os.path.isdir(subjects_dir):
        os.makedirs(subjects_dir)
    os.environ["SINGULARITYENV_SUBJECTS_DIR"]=subjects_dir

    command_base, container = getContainer(labels_dict,nodename="freesurfer", SPECIFIC="FREESURFER_CONTAINER",LOGGER=IFLOGGER)   
    FREEVER="Unknown"
    IFLOGGER.info("Checking the recon-all version:")
    command = f"{command_base} recon-all --version"
    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)
    if "-7.3.2-" in results:
        FREEVER="7.3.2"

    IFLOGGER.info("\nChecking the container version:")
    command = f"{command_base} --version"
    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    T1wfile=None
    T1wLabel = getParams(labels_dict,'T1W')
    t1w_entity={"acquisition": None, "reconstruction": None}
    if T1wLabel is not None:
        t1w_entity=add_labels(T1wLabel,t1w_entity)
        t1w_entity["subject"]=participant_label
        t1w_entity["extension"]='nii.gz'
        if session_label:
            t1w_entity["session"]=session_label
        T1w= layout.get(return_type='file', invalid_filters='allow', **t1w_entity)
        if T1w:
            T1wfile=T1w[0]
    else:
        T1w=layout.get(subject=participant_label,suffix='T1w', extension='nii.gz')
        if T1w:
            T1wfile=T1w[0].path

    HippoT2wfile=None
    HippoLabel = getParams(labels_dict,'HIPPO_T2W')
    hippo_t2w_entity={"acquisition": None, "reconstruction": None}
    if HippoLabel is not None:
        if "hippo_id" in HippoLabel.keys():
            hippo_id=HippoLabel["hippo_id"]
            HippoLabel.pop("hippo_id")
        else:
            hippo_id="T2w"
        hippo_t2w_entity=add_labels(HippoLabel,hippo_t2w_entity)
        hippo_t2w_entity["subject"]=participant_label
        hippo_t2w_entity["extension"]='nii.gz'
        if session_label:
            hippo_t2w_entity["session"]=session_label
        HippoT2w = layout.get(return_type='file', invalid_filters='allow', **hippo_t2w_entity)

        if HippoT2w:
            HippoT2wfile=HippoT2w[0]

    PialT2wfile=None
    PialLabel = getParams(labels_dict,'PIAL_T2W')
    pial_t2w_entity={"acquisition": None, "reconstruction": None}
    ISFLAIR=False
    if PialLabel is not None:
        pial_t2w_entity=add_labels(PialLabel,pial_t2w_entity)
        pial_t2w_entity["subject"]=participant_label
        pial_t2w_entity["extension"]='nii.gz'
        if session_label:
            pial_t2w_entity["session"]=session_label
        PialT2w = layout.get(invalid_filters='allow', **pial_t2w_entity)

        if PialT2w:
            PialT2wfile=PialT2w[0].path
            if PialT2w[0].entities['suffix'] == "FLAIR":
                ISFLAIR=True

    pial_t2_string = ""
    if PialT2wfile:
        if ISFLAIR:
            pial_t2_string = f" -FLAIR {PialT2wfile} -FLAIRpial"
            # https://github.com/freesurfer/freesurfer/issues/849
            # https://surfer.nmr.mgh.harvard.edu/fswiki/ReleaseNotes
            # Address issue with FLAIR-based pial reconstructions
            if FREEVER=="7.3.2":
                IFLOGGER.info("Due to known issue in Freesurfer 7.3.2 while using FLAIR images to reconstruct pial surface we will copy global-expert-options.txt to SUBJECTS_DIR")
                IFLOGGER.info("See https://github.com/freesurfer/freesurfer/issues/849 and https://surfer.nmr.mgh.harvard.edu/fswiki/ReleaseNotes")
                command=f"{command_base} cp /subjects_dir/global-expert-options.txt {subjects_dir}"
                evaluated_command=substitute_labels(command, labels_dict)
                IFLOGGER.info(evaluated_command)
                evaluated_command_args = shlex.split(evaluated_command)
                results = subprocess.run(evaluated_command_args,stdout=subprocess.PIPE,stderr=subprocess.STDOUT, text=True)
                IFLOGGER.info(results.stdout)

        else:
            pial_t2_string = f" -T2 {PialT2wfile} -T2pial"

    if not os.path.exists(os.path.join(subjects_dir,"sub-"+participant_label,"recon-all.log")):
        params="-all" \
            " -i " + T1wfile + \
            " -subject " + subject + \
            " -parallel" \
            + pial_t2_string + \
            " -openmp <BIDSAPP_THREADS>"

        command=f"{command_base} recon-all"\
                " "+params
    else:
        params="-all" \
            " -no-isrunning" \
            " -subject " + subject + \
            " -parallel" \
            + pial_t2_string + \
            " -openmp <BIDSAPP_THREADS>"

        command=f"{command_base} recon-all"\
                " "+params


    evaluated_command=substitute_labels(command,labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)


    # Hipposubfields segmentation just T1
    params= subject 

    command=f"{command_base} segmentHA_T1.sh"\
            " "+params

    evaluated_command=substitute_labels(command,labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    # Test new beta Hipposubfields segmentation released in 7.3.2 that only uses T1
    params= " hippo-amygdala"\
            " --cross " + subject + \
            " --suffix beta" 

    command=f"{command_base} segment_subregions"\
            " "+params

    evaluated_command=substitute_labels(command,labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)


    if HippoT2wfile:
        # Hipposubfields segmentation using T2 and T1
        useT1="1"
        params= subject + \
            " " + HippoT2wfile + \
            " " + hippo_id + \
            " " + useT1

        command=f"{command_base} segmentHA_T2.sh"\
                " "+params

        evaluated_command=substitute_labels(command,labels_dict)
        results = runCommand(evaluated_command,IFLOGGER)

        # Hipposubfields segmentation using just T2
        useT1="0"
        params= subject + \
            " " + HippoT2wfile + \
            " " + hippo_id + \
            " " + useT1

        command=f"{command_base} segmentHA_T2.sh"\
                " "+params

        evaluated_command=substitute_labels(command,labels_dict)
        results = runCommand(evaluated_command,IFLOGGER)
    else:
        hippo_id="T1"


    # Thalamic Segmentation 
    params= subject

    command=f"{command_base} segmentThalamicNuclei.sh"\
            " "+params

    evaluated_command=substitute_labels(command,labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    # Test new beta Thalamic and Brainstem segmentation released in 7.3.2 that only uses T1
    params= " thalamus"\
            " --cross " + subject + \
            " --suffix beta_thalamus" 

    command=f"{command_base} segment_subregions"\
            " "+params

    evaluated_command=substitute_labels(command,labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)


    params= " brainstem"\
            " --cross " + subject + \
            " --suffix beta_brainstem" 

    command=f"{command_base} segment_subregions"\
            " "+params

    evaluated_command=substitute_labels(command,labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

 
    L_hipposubfields = getGlob(os.path.join(subjects_dir,'sub-{}'.format(participant_label),'stats','hipposubfields.lh.*{}*'.format(hippo_id)))   
    R_hipposubfields = getGlob(os.path.join(subjects_dir,'sub-{}'.format(participant_label),'stats','hipposubfields.rh.*{}*'.format(hippo_id)))
    L_a2009stats = getGlob(os.path.join(subjects_dir,'sub-{}'.format(participant_label),'stats','lh.aparc.a2009s.stats'))
    R_a2009stats = getGlob(os.path.join(subjects_dir,'sub-{}'.format(participant_label),'stats','rh.aparc.a2009s.stats'))
    asegstats =  getGlob(os.path.join(subjects_dir,'sub-{}'.format(participant_label),'stats','aseg.stats'))
    output_dir = cwd

    
    out_files=[]
    out_files.insert(0,L_hipposubfields)
    out_files.insert(1,R_hipposubfields)
    out_files.insert(2,L_a2009stats)
    out_files.insert(3,R_a2009stats)
    out_files.insert(4,asegstats)


    return {
        "L_hipposubfields": L_hipposubfields,
        "R_hipposubfields": R_hipposubfields,
        "L_a2009stats": L_a2009stats,
        "R_a2009stats": R_a2009stats,
        "asegstats": asegstats,
        "subjects_dir": subjects_dir,
        "output_dir":output_dir,
        "out_files":out_files
    }



class freesurferInputSpec(BaseInterfaceInputSpec):
    labels_dict = traits.Dict({},mandatory=False,desc='labels', usedefault=True)
    bids_dir = traits.String("",desc="BIDS Directory", usedefault=True)

class freesurferOutputSpec(TraitedSpec):
    L_hipposubfields = File(desc='Hipposubfield stats LH')
    R_hipposubfields = File(desc='Hipposubfield stats RH')
    L_a2009stats = File(desc='aparc stats LH')
    R_a2009stats = File(desc='aparc stats RH')
    asegstats = File(desc='aseg stats')
    subjects_dir = traits.String(desc="Freesurfer subjects directory")
    output_dir = traits.String(desc="Freesurfer output directory")
    out_files = traits.List(desc='list of files')
    
class freesurfer_pan(BaseInterface):
    input_spec = freesurferInputSpec
    output_spec = freesurferOutputSpec

    def _run_interface(self, runtime):

        # Call our python code here:
        outputs = freesurfer_proc(
            self.inputs.labels_dict,
            self.inputs.bids_dir
        )

        setattr(self, "_results", outputs)
        # And we are done
        return runtime

    def _list_outputs(self):
        return self._results


def create(labels_dict,name='freesurfer_node',bids_dir="", LOGGER=IFLOGGER):
    # Create Node
    pan_node = Node(freesurfer_pan(), name=name)

    if LOGGER:
        LOGGER.info(f"Created Node {pan_node!r}")
        
    # Specify node inputs
    pan_node.inputs.labels_dict = labels_dict
    
    if bids_dir is None or bids_dir == "":
        bids_dir = substitute_labels("<BIDS_DIR>", labels_dict)

    pan_node.inputs.bids_dir =  bids_dir

    return pan_node





import nitransforms as nit
import nibabel as nib
import os
import importlib
import tempfile
import glob
from subprocess import check_call
import re
from panpipelines.utils import util_functions as ut
import logging
import sys

TRANSFORM_LITERALS=["identity"]

LOGGER = logging.getLogger("panpipelines.utils.transformer")
LOGGER.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(name)s | %(asctime)s | %(levelname)s | %(message)s')
stdout_handler = logging.StreamHandler(sys.stdout)
stdout_handler.setLevel(logging.INFO)
stdout_handler.setFormatter(formatter)
LOGGER.addHandler(stdout_handler)


def fsleyes_view(command_base,params):
    command=f"{command_base} fsleyes"\
            " "+params
    results = ut.runCommand(command)

def get_orientation_from_file(filename,type="default"):
    if type=="ants_transform":
        antstrans=nit.io.itk.ITKLinearTransform.from_filename(filename)
        affine=antstrans.to_ras()
    elif type=="image":
        img = nib.load(filename)
        affine=img.affine
    else:
        affine=filename
    return ["".join(nib.orientations.aff2axcodes(affine)),affine]

def reorient(input_file, ori, out_file=None):
    from nipype.interfaces.image import Reorient
    from nipype import Node

    reorient_node = Node(Reorient(),name="reorient_to_{}".format(ori))
    reorient_node.inputs.in_file =  input_file
    reorient_node.inputs.orientation=ori
    reorient_results = reorient_node.run()
    if out_file is not None:
        if os.path.isdir(out_file):
            command = "cp {} {}".format(reorient_results.outputs.out_file,out_file)
            results = ut.runCommand(command)
            out_file=os.path.join(out_file, os.path.basename(reorient_results.outputs.out_file))
        else:
            command = "cp {} {}".format(reorient_results.outputs.out_file,out_file)
            results = ut.runCommand(command)
    else:
        out_file = reorient_results.outputs.out_file
    return out_file

def convert_affine_ants_to_fsl(antstrans_file, moving, reference, fsltrans_file):
    antstrans=nit.io.itk.ITKLinearTransform.from_filename(antstrans_file)
    movimg=nib.load(moving)
    refimg=nib.load(reference)
    fsltr=nit.io.fsl.FSLLinearTransform.from_ras(antstrans.to_ras(),movimg,refimg)
    fsltr.to_filename(fsltrans_file)

def convert_affine_fsl_to_ants(fsltrans_file, moving, reference, antstrans_file):
    fsltrans=nit.io.fsl.FSLLinearTransform.from_filename(fsltrans_file)
    movimg=nib.load(moving)
    refimg=nib.load(reference)
    antstr=nit.io.itk.ITKLinearTransform.from_ras(fsltrans.to_ras(moving,refimg))
    antstr.to_filename(antstrans_file)

def get_template_ref(TEMPLATEFLOW_HOME,template_space,suffix=None,desc=None,resolution=None,extension=[".nii.gz"]):
    os.environ["TEMPLATEFLOW_HOME"]=TEMPLATEFLOW_HOME
    
    from templateflow import api as tf
    
    template_ref=tf.get(template_space,resolution=resolution,desc=desc,suffix=suffix,extension=extension)
    return template_ref

def apply_transform_mni_to_mni2009_ants_ori(TEMPLATEFLOW_HOME,input_file,out_file, COMMANDBASE,resolution=1,transform_ori="RAS:RAS",target_ori="RAS",reverse=False, costfunction=None,output_type=None):
    os.environ["TEMPLATEFLOW_HOME"]=TEMPLATEFLOW_HOME

    from templateflow import api as tf
    mninlin6_mni2009_trans=tf.get("MNI152NLin2009cAsym",suffix="xfm",extension=[".h5"])
    mni2009_t1_ref=tf.get("MNI152NLin2009cAsym",resolution=resolution,desc=None,suffix="T1w",extension=[".nii.gz"])

    apply_transform_ants_ori(input_file, mni2009_t1_ref,out_file,mninlin6_mni2009_trans,COMMANDBASE,transform_ori="RAS:RAS",target_ori="RAS",reverse=reverse, costfunction=costfunction,output_type=output_type)


def apply_transform_mni_to_mni2009_ants(TEMPLATEFLOW_HOME,input_file,out_file, COMMANDBASE,resolution=1,reverse=False, costfunction=None,output_type=None):
    os.environ["TEMPLATEFLOW_HOME"]=TEMPLATEFLOW_HOME

    from templateflow import api as tf
    mninlin6_mni2009_trans=tf.get("MNI152NLin2009cAsym",suffix="xfm",extension=[".h5"])
    mni2009_t1_ref=tf.get("MNI152NLin2009cAsym",resolution=resolution,desc=None,suffix="T1w",extension=[".nii.gz"])

    apply_transform_ants(input_file, mni2009_t1_ref,out_file,mninlin6_mni2009_trans,COMMANDBASE,reverse=reverse, costfunction=costfunction,output_type=output_type)

def apply_transform_ants_ori(input_file,ref_file, out_file, trans_mat, COMMANDBASE, transform_ori="RAS:RAS",target_ori="RAS", reverse=False, costfunction=None, output_type=None):

    input_file = os.path.abspath(input_file)
    ref_file=os.path.abspath(ref_file)
    out_file=os.path.abspath(out_file)

    expected_mov_ori = transform_ori.split(":")[0]
    expected_ref_ori = transform_ori.split(":")[1]
    
    # Ensure input_file, reference_file and transform are in the same orientation
    ref_ori = get_orientation_from_file(ref_file,"image")
    if not ref_ori[0] == expected_ref_ori:
        print("reorienting  ref_file {} from ref_ori {} to expected_ref_ori {}".format(ref_file, ref_ori, expected_ref_ori))
        ref_file=reorient(ref_file, expected_ref_ori, os.path.dirname(out_file))

    mov_ori = get_orientation_from_file(input_file,"image")
    if not mov_ori[0] == expected_mov_ori:
        print("reorienting input_file (moving) {} from mov_ori {} to expected_mov_ori {}".format(input_file, mov_ori, expected_mov_ori))
        input_file=reorient(input_file, expected_mov_ori, os.path.dirname(out_file))

    if costfunction is None:
        costfunction="LanczosWindowedSinc"

    img = nib.load(input_file)
    dimz=1
    if len(img.header.get_data_shape()) > 3:
        dimz = img.header.get_data_shape()[3]

    image_type = "1"
    if dimz > 1:
        image_type = "3"

    if output_type:
        output_type = f"-u {output_type}"
    else:
        output_type = ""

    TRANSFORMS=""
    if isinstance(trans_mat,list) and isinstance(reverse,list):
        for transcount in range(len(trans_mat)):
            if not trans_mat[transcount] in TRANSFORM_LITERALS:
                trans_item = os.path.abspath(trans_mat[transcount])
            else:
                trans_item = trans_mat[transcount]
            reverse_item = reverse[transcount]
            if reverse_item:
                TRANSFORMS = " -t [" + str(trans_item) + ",1]" + TRANSFORMS
            else:
                TRANSFORMS = " -t " + str(trans_item)  + TRANSFORMS
    else:
        trans_math = os.path.abspath(trans_mat)
        if reverse:
            TRANSFORMS = " -t [" + str(trans_mat) + ",1]"
        else:
            TRANSFORMS = " -t " + str(trans_mat) 
    # for labels, atlases consider -n NearestNeighbor
    params="-d 3" \
        " -e " + image_type +\
        " -i " + str(input_file) +\
        " -f 0"\
        " --float 1"\
        " -n " + costfunction+\
        " -o " + str(out_file) +\
        " -r " + str(ref_file) + \
        TRANSFORMS + \
        " -v 1"  + \
        " " + output_type

    command=f"{COMMANDBASE} antsApplyTransforms"\
            " "+params

    print(command)
    results = ut.runCommand(command)

    # quick hack to fix issue with templateflow transfomr - 5 dims instead of 3 dims used in header
    command=f"{COMMANDBASE} fslroi"\
            " "+out_file+" "+out_file + " 0 "+str(dimz)
    results = ut.runCommand(command)

    # Transform to target orientation
    # Ensure input_file, reference_file and transform are in the same orientation
    actual_target_ori = get_orientation_from_file(out_file,"image")
    if not actual_target_ori[0] == target_ori:
        print("reorienting  target_file {} from actual_target_ori {} to {}".format(out_file, actual_target_ori, target_ori))
        out_file=reorient(out_file, target_ori,out_file)

    return out_file


def apply_transform_ants(input_file,ref_file, out_file, trans_mat, COMMANDBASE, reverse=False, costfunction=None,output_type=None):

    input_file = os.path.abspath(input_file)
    ref_file=os.path.abspath(ref_file)
    out_file=os.path.abspath(out_file)
    
    if costfunction is None:
        costfunction="LanczosWindowedSinc"

    img = nib.load(input_file)
    dimz=1
    if len(img.header.get_data_shape()) > 3:
        dimz = img.header.get_data_shape()[3]

    image_type = "1"
    if dimz > 1:
        image_type = "3"

    if output_type:
        output_type = f"-u {output_type}"
    else:
        output_type = ""

    TRANSFORMS=""
    if isinstance(trans_mat,list) and isinstance(reverse,list):
        for transcount in range(trans_mat):
            trans_item = os.path.abspath(trans_mat[transcount])
            reverse_item = reverse[transcount]
            if reverse_item:
                TRANSFORMS = " -t [" + str(trans_item) + ",1]" + TRANSFORMS
            else:
                TRANSFORMS = " -t" + str(trans_item)  + TRANSFORMS
    else:
        trans_mat=os.path.abspath(trans_mat)
        if reverse:
            TRANSFORMS = " -t [" + trans_mat + ",1]"
        else:
            TRANSFORMS = " -t " + str(trans_mat) 

    # for labels, atlases consider -n NearestNeighbor
    params="-d 3" \
        " -e " + image_type +\
        " -i " + str(input_file) +\
        " -f 0"\
        " --float 1"\
        " -n " + costfunction+\
        " -o " + str(out_file) +\
        " -r " + str(ref_file) + \
        TRANSFORMS + \
        " -v 1" + \
        " " + output_type

    command=f"{COMMANDBASE} antsApplyTransforms"\
            " "+params

    results = ut.runCommand(command)

    # quick hack to fix issue with templateflow transfomr - 5 dims instead of 3 dims used in header
    command=f"{COMMANDBASE} fslroi"\
            " "+out_file+" "+out_file + " 0 "+str(dimz)
    results = ut.runCommand(command)


def resampleimage_ants_ori(input_file, out_file, newdims, COMMANDBASE, dim_type="0",target_ori=None,interpolation_type=None,output_type=None):

    input_file = os.path.abspath(input_file)
    out_file=os.path.abspath(out_file)

    if not dim_type:
        # use default = spacing
        dim_type="0"
    
    img = nib.load(input_file)
    dimz=1
    image_dim = "3"
    if len(img.header.get_data_shape()) > 3:
        dimz = img.header.get_data_shape()[3]
        image_dim = "4"
        newdims = newdims + f"x{dimz}"

    if not interpolation_type:
        # use default = linear
        interpolation_type=""

    if not output_type:
        # use detault = float
        output_type = ""

    params=f"{image_dim}" \
        f" {input_file}"\
        f" {out_file}"\
        f" {newdims}"\
        f" {dim_type}"\
        f" {interpolation_type}"\
        f" {output_type}"

    command=f"{COMMANDBASE} ResampleImage"\
            " "+params

    results = ut.runCommand(command)
    # Transform to target orientation
    # Ensure input_file, reference_file and transform are in the same orientation
    if target_ori:
        actual_target_ori = get_orientation_from_file(out_file,"image")
        if not actual_target_ori[0] == target_ori:
            print("reorienting  target_file {} from actual_target_ori {} to {}".format(out_file, actual_target_ori, target_ori))
            out_file=reorient(out_file, target_ori,out_file)

    return out_file


def resample_ants_ori(input_file,ref_file, out_file, COMMANDBASE,transform_ori="RAS:RAS",target_ori="RAS", costfunction=None,output_type=None):

    input_file = os.path.abspath(input_file)
    ref_file=os.path.abspath(ref_file)
    out_file=os.path.abspath(out_file)
    
    expected_mov_ori = transform_ori.split(":")[0]
    expected_ref_ori = transform_ori.split(":")[1]
    
    # Ensure input_file, reference_file and transform are in the same orientation
    ref_ori = get_orientation_from_file(ref_file,"image")
    if not ref_ori[0] == expected_ref_ori:
        print("reorienting  ref_file {} from ref_ori {} to expected_ref_ori {}".format(ref_file, ref_ori, expected_ref_ori))
        ref_file=reorient(ref_file, expected_ref_ori, os.path.dirname(out_file))

    mov_ori = get_orientation_from_file(input_file,"image")
    if not mov_ori[0] == expected_mov_ori:
        print("reorienting input_file (moving) {} from mov_ori {} to expected_mov_ori {}".format(input_file, mov_ori, expected_mov_ori))
        input_file=reorient(input_file, expected_mov_ori, os.path.dirname(out_file))

    img = nib.load(input_file)
    dimz=1
    if len(img.header.get_data_shape()) > 3:
        dimz = img.header.get_data_shape()[3]

    image_type = "1"
    if dimz > 1:
        image_type = "3"

    if costfunction is None:
        costfunction="LanczosWindowedSinc"

    if output_type:
        output_type = f"-u {output_type}"
    else:
        output_type = ""

    params="-d 3" \
        " -e " + image_type +\
        " -i " + str(input_file) +\
        " -f 0"\
        " --float 1"\
        " -o " + str(out_file) +\
        " -n " + costfunction+\
        " -r " + str(ref_file) + \
        " -t identity" + \
        " -v 1" + \
        " " + output_type

    command=f"{COMMANDBASE} antsApplyTransforms"\
            " "+params

    results = ut.runCommand(command)

    # Transform to target orientation
    # Ensure input_file, reference_file and transform are in the same orientation
    actual_target_ori = get_orientation_from_file(out_file,"image")
    if not actual_target_ori[0] == target_ori:
        print("reorienting  target_file {} from actual_target_ori {} to {}".format(out_file, actual_target_ori, target_ori))
        out_file=reorient(out_file, target_ori,out_file)

    return out_file
    
def resample_ants(input_file,ref_file, out_file, COMMANDBASE, costfunction=None,output_type=None):

    input_file = os.path.abspath(input_file)
    ref_file=os.path.abspath(ref_file)
    out_file=os.path.abspath(out_file)
    
    img = nib.load(input_file)
    dimz=1
    if len(img.header.get_data_shape()) > 3:
        dimz = img.header.get_data_shape()[3]

    image_type = "1"
    if dimz > 1:
        image_type = "3"

    if costfunction is None:
        costfunction="LanczosWindowedSinc"

    if output_type:
        output_type = f"-u {output_type}"
    else:
        output_type = ""

    params="-d 3" \
        " -e " + image_type +\
        " -i " + str(input_file) +\
        " -f 0"\
        " --float 1"\
        " -o " + str(out_file) +\
        " -n " + costfunction+\
        " -r " + str(ref_file) + \
        " -t identity" + \
        " -v 1" + \
        " " + output_type

    command=f"{COMMANDBASE} antsApplyTransforms"\
            " "+params

    results = ut.runCommand(command)

def ants_registration_ori(moving, reference, transform, COMMANDBASE, transform_ori="RAS:RAS",target_ori="RAS", composite=False):
    
    transform=os.path.abspath(transform)
    reference=os.path.abspath(reference)
    moving=os.path.abspath(moving)
    
    expected_mov_ori = transform_ori.split(":")[0]
    expected_ref_ori = transform_ori.split(":")[1]
    
    # Ensure input_file, reference_file and transform are in the same orientation
    ref_ori = get_orientation_from_file(reference,"image")
    if not ref_ori[0] == expected_ref_ori:
        print("reorienting  ref_file {} from ref_ori {} to expected_ref_ori {}".format(reference, ref_ori, expected_ref_ori))
        reference=reorient(reference, expected_ref_ori, os.path.dirname(transform))

    mov_ori = get_orientation_from_file(moving,"image")
    if not mov_ori[0] == expected_mov_ori:
        print("reorienting input_file (moving) {} from mov_ori {} to expected_mov_ori {}".format(moving, mov_ori, expected_mov_ori))
        moving=reorient(moving, expected_mov_ori, os.path.dirname(transform))
    
    if composite:
        composite_param= " --write-composite-transform 1"
    else:
        composite_param= " --write-composite-transform 0"

    params="--collapse-output-transforms 1" \
    " --dimensionality 3" \
    " --winsorize-image-intensities [ 0.025, 0.975 ]"\
    " --use-histogram-matching 1"\
    f" --initial-moving-transform [ {reference}, {moving}, 1 ] "\
    " --initialize-transforms-per-stage 0" \
    " --interpolation LanczosWindowedSinc" \
    f" --output [ {transform}, {transform}_Warped.nii.gz ]"\
    " --transform Rigid[ 0.1 ] "\
    f" --MI[ {reference}, {moving}, 1, 32, Regular, 0.25 ]"\
    " --convergence [ 1000x500x250x100, 1e-06, 10 ]"\
    " --shrink-factors 8x4x2x1"\
    " --smoothing-sigmas 3x2x1x0vox"\
    " --transform Affine[ 0.1 ] "\
    f" --MI[ {reference}, {moving}, 1, 32, Regular, 0.25 ]"\
    " --convergence [ 1000x500x250x100, 1e-06, 10 ]"\
    " --shrink-factors 8x4x2x1"\
    " --smoothing-sigmas 3x2x1x0vox"\
    " --transform SyN[0.1,3,0] "\
    f" --CC[ {reference}, {moving}, 1,4 ]"\
    " --convergence [ 100x70x50x20, 1e-06, 10 ]"\
    " --shrink-factors 8x4x2x1"\
    " --smoothing-sigmas 3x2x1x0vox"\
    + composite_param

    
    command=f"{COMMANDBASE} antsRegistration"\
            " "+params

    results = ut.runCommand(command)
    
    transform_dir=os.path.dirname(transform)
    dirs = os.listdir(transform_dir)
    candidates=[s for s in dirs if re.findall(r'.*' +os.path.basename(transform)+ '\d.*',s)]
    forward = [os.path.join(transform_dir,s) for s in candidates if "Inverse".upper() not in s.upper()]
    inverse = [os.path.join(transform_dir,s) for s in candidates if "Inverse".upper() in s.upper()]
                             
    forward.sort()

    return {"forward": forward, "inverse" : inverse}
    
    
    
def ants_registration(moving, reference, transform, COMMANDBASE, composite=False):
    
    transform=os.path.abspath(transform)
    reference=os.path.abspath(reference)
    moving=os.path.abspath(moving)
    
    if composite:
        composite_param= " --write-composite-transform 1"
    else:
        composite_param= " --write-composite-transform 0"

    params="--collapse-output-transforms 1" \
    " --dimensionality 3" \
    " --winsorize-image-intensities [ 0.025, 0.975 ]"\
    " --use-histogram-matching 1"\
    f" --initial-moving-transform [ {reference}, {moving}, 1 ] "\
    " --initialize-transforms-per-stage 0" \
    " --interpolation LanczosWindowedSinc" \
    f" --output [ {transform}, {transform}_Warped.nii.gz ]"\
    " --transform Rigid[ 0.1 ] "\
    f" --metric MI[{reference}, {moving}, 1, 32, Regular, 0.25 ]"\
    " --convergence [ 1000x500x250x100, 1e-06, 10 ]"\
    " --shrink-factors 8x4x2x1"\
    " --smoothing-sigmas 3x2x1x0vox"\
    " --transform Affine[ 0.1 ] "\
    f" --metric MI[{reference}, {moving}, 1, 32, Regular, 0.25 ]"\
    " --convergence [ 1000x500x250x100, 1e-06, 10 ]"\
    " --shrink-factors 8x4x2x1"\
    " --smoothing-sigmas 3x2x1x0vox"\
    " --transform SyN[0.1,3,0] "\
    f" --metric CC[{reference}, {moving}, 1,4 ]"\
    " --convergence [ 100x70x50x20, 1e-06, 10 ]"\
    " --shrink-factors 8x4x2x1"\
    " --smoothing-sigmas 3x2x1x0vox"\
    + composite_param

    
    command=f"{COMMANDBASE} antsRegistration"\
            " "+params

    results = ut.runCommand(command)
    
    transform_dir=os.path.dirname(transform)
    dirs = os.listdir(transform_dir)
    candidates=[s for s in dirs if re.findall(r'.*' +os.path.basename(transform)+ '\d.*',s)]
    forward = [os.path.join(transform_dir,s) for s in candidates if "Inverse".upper() not in s.upper()]
    inverse = [os.path.join(transform_dir,s) for s in candidates if "Inverse".upper() in s.upper()]
                             
    forward.sort()

    return {"forward": forward, "inverse" : inverse}
    

# ensure that moving and reference are in RAS before calculating transform
def ants_registration_rigid_ori(moving, reference, transform, COMMANDBASE,transform_ori="RAS:RAS",target_ori="RAS"):
    
    transform=os.path.abspath(transform)
    reference=os.path.abspath(reference)
    moving=os.path.abspath(moving)
       
    expected_mov_ori = transform_ori.split(":")[0]
    expected_ref_ori = transform_ori.split(":")[1]
    
    # Ensure input_file, reference_file and transform are in the same orientation
    ref_ori = get_orientation_from_file(reference,"image")
    if not ref_ori[0] == expected_ref_ori:
        print("reorienting  ref_file {} from ref_ori {} to expected_ref_ori {}".format(reference, ref_ori, expected_ref_ori))
        reference=reorient(reference, expected_ref_ori, os.path.dirname(transform))

    mov_ori = get_orientation_from_file(moving,"image")
    if not mov_ori[0] == expected_mov_ori:
        print("reorienting input_file (moving) {} from mov_ori {} to expected_mov_ori {}".format(moving, mov_ori, expected_mov_ori))
        moving=reorient(moving, expected_mov_ori, os.path.dirname(transform))

    params="--collapse-output-transforms 1" \
    " --dimensionality 3" \
    f" --initial-moving-transform [ {reference}, {moving}, 1 ] "\
    " --initialize-transforms-per-stage 0" \
    " --interpolation LanczosWindowedSinc" \
    f" --output [ {transform}, {transform}_Warped.nii.gz ]"\
    " --transform Rigid[ 0.2 ] "\
    f" --metric Mattes[ {reference}, {moving}, 1, 32, Random, 0.25 ]"\
    " --convergence [ 10000x1000x10000x10000, 1e-06, 10 ]"\
    " --smoothing-sigmas 7.0x3.0x1.0x0.0vox"\
    " --shrink-factors 8x4x2x1"\
    " --use-histogram-matching 1"\
    " --winsorize-image-intensities [ 0.025, 0.975 ]"\
    " --write-composite-transform 0"

    command=f"{COMMANDBASE} antsRegistration"\
            " "+params

    results = ut.runCommand(command)
    
    final_transform=transform+"0GenericAffine.mat"

    return final_transform

def ants_registration_rigid(moving, reference, transform, COMMANDBASE):
    
    transform=os.path.abspath(transform)
    reference=os.path.abspath(reference)
    moving=os.path.abspath(moving)

    params="--collapse-output-transforms 1" \
    " --dimensionality 3" \
    f" --initial-moving-transform [ {reference}, {moving}, 1 ] "\
    " --initialize-transforms-per-stage 0" \
    " --interpolation LanczosWindowedSinc" \
    f" --output [ {transform}, {transform}_Warped.nii.gz ]"\
    " --transform Rigid[ 0.2 ] "\
    f" --metric Mattes[ {reference}, {moving}, 1, 32, Random, 0.25 ]"\
    " --convergence [ 10000x1000x10000x10000, 1e-06, 10 ]"\
    " --smoothing-sigmas 7.0x3.0x1.0x0.0vox"\
    " --shrink-factors 8x4x2x1"\
    " --use-histogram-matching 1"\
    " --winsorize-image-intensities [ 0.025, 0.975 ]"\
    " --write-composite-transform 0"

    command=f"{COMMANDBASE} antsRegistration"\
            " "+params

    results = ut.runCommand(command)
    
    final_transform=transform+"0GenericAffine.mat"

    return final_transform

# ensure that moving and reference are in RAS before calculating transform
def ants_registration_quick_ori(moving, reference, transform, COMMANDBASE,transform_ori="RAS:RAS",target_ori="RAS",threads=8):
    
    transform=os.path.abspath(transform)
    reference=os.path.abspath(reference)
    moving=os.path.abspath(moving)
       
    if threads is not None:
        threadparams=" -n " + str(threads)
        
    expected_mov_ori = transform_ori.split(":")[0]
    expected_ref_ori = transform_ori.split(":")[1]
    
    # Ensure input_file, reference_file and transform are in the same orientation
    ref_ori = get_orientation_from_file(reference,"image")
    if not ref_ori[0] == expected_ref_ori:
        print("reorienting  ref_file {} from ref_ori {} to expected_ref_ori {}".format(reference, ref_ori, expected_ref_ori))
        reference=reorient(reference, expected_ref_ori, os.path.dirname(transform))

    mov_ori = get_orientation_from_file(moving,"image")
    if not mov_ori[0] == expected_mov_ori:
        print("reorienting input_file (moving) {} from mov_ori {} to expected_mov_ori {}".format(moving, mov_ori, expected_mov_ori))
        moving=reorient(moving, expected_mov_ori, os.path.dirname(transform))

    params=f" -f {reference}" \
    f" -m {moving}" \
    f" -o {transform}"\
    + threadparams

    command=f"{COMMANDBASE} antsRegistrationSyNQuick.sh"\
            " "+params

    results = ut.runCommand(command)
    
    transform_dir=os.path.dirname(transform)
    dirs = os.listdir(transform_dir)
    candidates=[s for s in dirs if re.findall(r'.*' +os.path.basename(transform)+ '\d.*',s)]
    forward = [os.path.join(transform_dir,s) for s in candidates if "Inverse".upper() not in s.upper()]
    inverse = [os.path.join(transform_dir,s) for s in candidates if "Inverse".upper() in s.upper()]
                             
    forward.sort()

    return {"forward": forward, "inverse" : inverse}

# ensure that moving and reference are in RAS before calculating transform
def ants_registration_quick(moving, reference, transform, COMMANDBASE,threads=8):
    
    transform=os.path.abspath(transform)
    reference=os.path.abspath(reference)
    moving=os.path.abspath(moving)
       
    if threads is not None:
        threadparams=" -n " + str(threads)
        
    params=f" -f {reference}" \
    f" -m {moving}" \
    f" -o {transform}"\
    + threadparams

    command=f"{COMMANDBASE} antsRegistrationSyNQuick.sh"\
            " "+params

    results = ut.runCommand(command)
    
    transform_dir=os.path.dirname(transform)
    dirs = os.listdir(transform_dir)
    candidates=[s for s in dirs if re.findall(r'.*' +os.path.basename(transform)+ '\d.*',s)]
    forward = [os.path.join(transform_dir,s) for s in candidates if "Inverse".upper() not in s.upper()]
    inverse = [os.path.join(transform_dir,s) for s in candidates if "Inverse".upper() in s.upper()]
                             
    forward.sort()

    return {"forward": forward, "inverse" : inverse}


    return candidates


def fsl_reg_flirt(input,reference,out,transmat,COMMANDBASE,dof="12",cost="mutualinfo"):
    command=f"{COMMANDBASE} flirt"\
        f" -in {input}"\
        f" -ref {reference}"\
        f" -out {out}"\
        f" -omat {transmat}"\
        f" -dof {dof}"\
        f" -cost {cost}"
    results = ut.runCommand(command)

def applyAffine_flirt(input,reference,out,transmat,COMMANDBASE,interp="trilinear"):
    command=f"{COMMANDBASE} flirt"\
        f" -in {input}"\
        f" -ref {reference}"\
        f" -out {out}"\
        f" -applyxfm"\
        f" -interp {interp}"\
        f" -init {transmat}"
    results = ut.runCommand(command)

def fsl_reg_fnirt(input,reference,out,transwarp,COMMANDBASE,transmat=None):
    if transmat is not None:
        transmat_param=f" --aff={transmat}"
    else:
        transmat_param=""
    
    command=f"{COMMANDBASE} fnirt"\
        f" --in={input}"\
        f" --ref={reference}"\
        f" --iout={out}"\
        + transmat_param + \
        f" --cout={transwarp}"
    
    results = ut.runCommand(command)
    
def fsl_register_nonlin(input,reference,out,COMMANDBASE):
    trans_str= ut.getTransName(input, reference)
    pre_aff = ut.newfile(prefix=f"{trans_str}",suffix="affine",extension="mat")
    pre_aff_out = ut.newfile(assocfile=pre_aff,suffix="outfile",extension=".nii.gz")
    fsl_reg_flirt(input,reference,pre_aff_out,pre_aff,COMMANDBASE)
    
    fwd_warp = ut.newfile(prefix=f"{trans_str}",suffix="warp",extension="nii.gz")
    fwd_warp_out = ut.newfile(assocfile=fwd_warp,suffix="outfile",extension=".nii.gz")
    fsl_reg_fnirt(input,reference,fwd_warp_out,fwd_warp,COMMANDBASE,transmat=pre_aff)
    
    inv_warp = ut.newfile(prefix=f"{trans_str}",suffix="invwarp",extension="nii.gz")
    invertWarpfield_FNIRT(input, fwd_warp, inv_warp, COMMANDBASE)
    
    return {"forward": fwd_warp, "inverse": inv_warp}

    
def applyWarp_fnirt(input,reference,out,transwarp,COMMANDBASE,interp="trilinear",premat=None,postmat=None):

    if premat is not None:
        premat_param=f"--premat={premat}"
    else:
        premat_param=""
        
    if postmat is not None:
        postmat_param=f"--postmat={postmat}"
    else:
        post_param=""
    
    command=f"{COMMANDBASE} applywarp"\
        f" --in={input}"\
        f" --ref={reference}"\
        f" --out={out}"\
        f" --warp={transwarp}"\
        f" --interp={interp}"\
        + premat_param + \
        + postmat_param 
    results = ut.runCommand(command)

def tkregister2_fslout(moving,target,COMMANDBASE,outmat_fsl):
    command=f"{COMMANDBASE} tkregister2"\
        f" --mov {moving}"\
        f" --targ {target}"\
        f" --regheader --reg junk --fslregout {outmat_fsl} --noedit"
    results = ut.runCommand(command)
    
def convMGZ2NII(mgz_in, nifti_out, COMMANDBASE):
    command=f"{COMMANDBASE} mri_convert"\
        f" --in_type mgz"\
        f" --out_type nii"\
        f" {mgz_in} {nifti_out}"
    results = ut.runCommand(command)
    
def disassembleTransforms(trans_in, trans_prefix, COMMANDBASE):
    command=f"{COMMANDBASE} --homedir={os.getcwd()} CompositeTransformUtil --disassemble"\
        f" {trans_in}"\
        f" {trans_prefix}"
    results = ut.runCommand(command)
    mats=glob.glob(os.path.join(os.getcwd(),f"*_{trans_prefix}_*"))
    return mats

def assembleTransforms(trans_in, trans_filename, COMMANDBASE):
    if isinstance(trans_in,list):
        trans_string = " ".join(trans_in)
        command=f"singularity run {COMMANDBASE} --homedir={os.getcwd()} CompositeTransformUtil --assemble"\
            f" {trans_filename}"\
            f" {trans_string}"
        results = ut.runCommand(command)
    else:
        print("list of transforms not passed. Skipping.")
        
def convertWarp_toFNIRT(ants_warp_field,fnirt_warp_field,source, COMMANDBASE ):
    command=f"{COMMANDBASE} wb_command -convert-warpfield"\
        f" -from-itk {ants_warp_field}"\
        f" -to-fnirt {fnirt_warp_field}"\
        f" {source}"   
    results = ut.runCommand(command)

def convertwarp_toANTS(fnirt_warp_field,source, ants_warp_field, COMMANDBASE,absolute="" ):
    command=f"{COMMANDBASE} wb_command -convert-warpfield"\
        f" -from-fnirt {fnirt_warp_field}"\
        f" {source}"\
        f" {absolute}"\
        f" -to-itk {ants_warp_field}"
    results = ut.runCommand(command)

def invertWarpfield_FNIRT(orig_source, fwd_warp_field, inv_warp_field, COMMANDBASE):
    command=f"{COMMANDBASE} invwarp"\
        f" --ref={orig_source}"\
        f" --warp={fwd_warp_field}"\
        f" --out={inv_warp_field}"
    results = ut.runCommand(command)
    
# absolute="-absolute" definitely did not work
def invertWarpfield_ANTS(ants_fwd_field, ants_inv_field, orig_source,orig_destination, COMMANDBASE,absolute=""):
    fnirt_fwd_field = tempfile.mkstemp()[1] + ".nii.gz"
    convertWarp_toFNIRT(ants_fwd_field,fnirt_fwd_field,orig_source, COMMANDBASE )
    
    fnirt_inv_field = tempfile.mkstemp()[1] + ".nii.gz"
    invertWarpfield_FNIRT(orig_source, fnirt_fwd_field, fnirt_inv_field , COMMANDBASE)
    
    convertwarp_toANTS(fnirt_inv_field,orig_destination, ants_inv_field, COMMANDBASE,absolute)

def invertAffine_FLIRT(fwd_affine, inv_affine, COMMANDBASE):
    command=f"{COMMANDBASE} comvert_xfm"\
        f" -omat {inv_affine}"\
        f" -inverse"\
        f" {fwd_affine}"
    results = ut.runCommand(command)

def invertAffine_ANTS(fwd_affine, inv_affine, moving, reference, COMMANDBASE):
    fsl_fwd_affine = tempfile.mkstemp()[1] + ".mat"
    convert_affine_ants_to_fsl(fwd_affine, moving, reference,  fsl_fwd_affine)
    
    fsl_inv_affine = tempfile.mkstemp()[1] + ".mat"
    invertAffine_FLIRT(fsl_fwd_affine, fsl_inv_affine, COMMANDBASE)
    
    convert_affine_fsl_to_ants(fsl_inv_affine, reference, moving, inv_affine)

if __name__ == '__main__':
    import sys
    if len(sys.argv) > 1:
        func=sys.argv[1]
        if func == "affine_ants_to_fsl":
            if len(sys.argv) > 5:
                antstrans_file=sys.argv[2]
                moving=sys.argv[3]
                refimg=sys.argv[4]
                fsltrans_file=sys.argv[5]
                convert_affine_ants_to_fsl(antstrans_file, moving, refimg, fsltrans_file)
        elif func == "ants_to_fsl":
            if len(sys.argv) > 5:
                antstrans_file=sys.argv[2]
                moving=sys.argv[3]
                refimg=sys.argv[4]
                fsltrans_file=sys.argv[5]
                convert_affine_ants_to_fsl(antstrans_file, moving, refimg, fsltrans_file)




from nipype.interfaces.base import BaseInterfaceInputSpec, BaseInterface, File, TraitedSpec, traits
from nipype import Node
from panpipelines.utils.util_functions import *
import os
import glob
import numpy as np 
import nibabel as nib
from nipype import logging as nlogging

IFLOGGER=nlogging.getLogger('nipype.interface')

def collate_csv_group_proc(labels_dict, csv_list1,csv_list2, add_prefix):

    if csv_list1 is None:
        csv_list1 = []
    if csv_list2 is None:
        csv_list2 = []
        
    cwd=os.getcwd()
    labels_dict = updateParams(labels_dict,"CWD",cwd)
    
    output_dir=cwd
    participants_label = getParams(labels_dict,'GROUP_PARTICIPANTS_LABEL')
    participants_project = getParams(labels_dict,'GROUP_PARTICIPANTS_XNAT_PROJECT')
    participants_session = getParams(labels_dict,'GROUP_SESSION_LABEL')
    pipeline = getParams(labels_dict,'PIPELINE')

    csv_list=[]

    if (participants_label is not None and (isinstance(participants_label,list) and len(participants_label) > 1)) and (participants_project is not None and (isinstance(participants_project,list) and len(participants_project)> 1)):
        for part_vals in zip(participants_label,participants_project,participants_session):
            labels_dict = updateParams(labels_dict,"PARTICIPANT_LABEL",part_vals[0])
            labels_dict = updateParams(labels_dict,"PARTICIPANT_XNAT_PROJECT",part_vals[1])
            labels_dict = updateParams(labels_dict,"PARTICIPANT_SESSION",part_vals[2])
            for meas_template in csv_list1:
                evaluated_meas_template = substitute_labels(meas_template,labels_dict)
                csv_list.extend(glob.glob(evaluated_meas_template))
            for meas_template in csv_list2:
                evaluated_meas_template = substitute_labels(meas_template,labels_dict)
                csv_list.extend(glob.glob(evaluated_meas_template)) 
    else:
        labels_dict = updateParams(labels_dict,"PARTICIPANT_LABEL","*")
        labels_dict = updateParams(labels_dict,"PARTICIPANT_XNAT_PROJECT","*")
        labels_dict = updateParams(labels_dict,"PARTICIPANT_SESSION","*")
        for meas_template in csv_list1:
            evaluated_meas_template = substitute_labels(meas_template,labels_dict)
            csv_list.extend(glob.glob(evaluated_meas_template))
        for meas_template in csv_list2:
            evaluated_meas_template = substitute_labels(meas_template,labels_dict)
            csv_list.extend(glob.glob(evaluated_meas_template))

    csv_list.sort()
    
    roi_output_dir = os.path.join(cwd,'group_roi_output_dir')
    if not os.path.isdir(roi_output_dir):
        os.makedirs(roi_output_dir)

    IFLOGGER.info(f"List of csv files to collate: {csv_list}")

    out_files=[]
    roi_csv_inner = None
    roi_csv_outer = None
    cum_df_inner=pd.DataFrame()
    cum_df_outer=pd.DataFrame()
    if len(csv_list) > 0:
        for csv_file in csv_list:
            df = pd.read_table(csv_file,sep=",")
            if cum_df_inner.empty:
                cum_df_inner = df
            else:
                cum_df_inner = pd.concat([cum_df_inner,df],join="inner")

            if cum_df_outer.empty:
                cum_df_outer = df
            else:
                cum_df_outer = pd.concat([cum_df_outer,df],join="outer")

        collate_name = getParams(labels_dict,"COLLATE_NAME")
        if not collate_name:
            if not pipeline:
                collate_name="csvgroup"
            else:
                collate_name="pipeline"

        roi_csv_inner = os.path.join(roi_output_dir,'{}_{}_inner.csv'.format("group",collate_name))
        roi_csv_outer = os.path.join(roi_output_dir,'{}_{}_outer.csv'.format("group",collate_name))

        cum_df_inner.to_csv(roi_csv_inner,sep=",",header=True, index=False)
        cum_df_outer.to_csv(roi_csv_outer,sep=",",header=True, index=False)

        out_files.insert(0,roi_csv_inner)
        out_files.insert(0,roi_csv_outer)

        metadata = {}
        roi_csv_inner_json = os.path.splitext(roi_csv_inner)[0] + ".json"
        metadata = updateParams(metadata,"Title","collate_csv_group.py: Inner Join")
        metadata = updateParams(metadata,"Description","Combine csv files of provided participants into group table. Only matched columns are retained.")
        metadata = updateParams(metadata,"MetadataFile",f"{roi_csv_inner_json}")
        metadata = updateParams(metadata,"FileCreated",f"{roi_csv_inner}")
        metadata = updateParams(metadata,"DateCreated",datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S:%f"))
        metadata = updateParams(metadata,"Pipeline",f"{pipeline}")
        metadata = updateParams(metadata,"InputFiles",f"{csv_list}")
        export_labels(metadata,roi_csv_inner_json)

        metadata = {}
        roi_csv_outer_json = os.path.splitext(roi_csv_outer)[0] + ".json"
        metadata = updateParams(metadata,"Title","collate_csv_group.py: Outer Join")
        metadata = updateParams(metadata,"Description","Combine csv files of provided participants into group table. Unmatched columns are retained.")
        metadata = updateParams(metadata,"MetadataFile",f"{roi_csv_outer_json}")
        metadata = updateParams(metadata,"FileCreated",f"{roi_csv_outer}")
        metadata = updateParams(metadata,"DateCreated",datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S:%f"))
        metadata = updateParams(metadata,"Pipeline",f"{pipeline}")
        metadata = updateParams(metadata,"InputFiles",f"{csv_list}")
        export_labels(metadata,roi_csv_outer_json)

    return {
        "roi_csv_inner":roi_csv_inner,
        "roi_csv_outer":roi_csv_outer,
        "roi_output_dir":roi_output_dir,
        "output_dir":output_dir,
        "out_files":out_files
    }


class collate_csv_groupInputSpec(BaseInterfaceInputSpec):
    labels_dict = traits.Dict({},mandatory=False,desc='labels', usedefault=True)
    csv_list1 = traits.List(desc='list of files')
    csv_list2 = traits.List(desc='list of files')
    add_prefix = traits.Bool(False,desc="Create header prefix while joining tables",usedefault=True)

class collate_csv_groupOutputSpec(TraitedSpec):
    roi_csv_inner = File(desc='CSV file of results')
    roi_csv_outer = File(desc='CSV file of results')
    roi_output_dir = traits.String(desc='roi output dir')
    output_dir = traits.String(desc='output dir')
    out_files = traits.List(desc='list of files')
    
class collate_csv_group_pan(BaseInterface):
    input_spec = collate_csv_groupInputSpec
    output_spec = collate_csv_groupOutputSpec

    def _run_interface(self, runtime):

        # Call our python code here:
        outputs = collate_csv_group_proc(
            self.inputs.labels_dict,
            self.inputs.csv_list1,
            self.inputs.csv_list2,
            self.inputs.add_prefix
        )

        setattr(self, "_results", outputs)
        # And we are done
        return runtime

    def _list_outputs(self):
        return self._results


def create(labels_dict,name="collate_csv_group_node",csv_list1="",csv_list2="",add_prefix=False,LOGGER=IFLOGGER):
    # Create Node
    pan_node = Node(collate_csv_group_pan(), name=name)

    if LOGGER:
        LOGGER.info(f"Created Node {pan_node!r}")

    # Specify node inputs
    pan_node.inputs.labels_dict = labels_dict

    if not csv_list1 is None and not csv_list1 == "":
        pan_node.inputs.csv_list1 = csv_list1
 
    if not csv_list2 is None and not csv_list2 == "":
        pan_node.inputs.csv_list2 = csv_list2

    pan_node.inputs.add_prefix =  add_prefix

    return pan_node





from nipype import Workflow, MapNode, Node

import panpipelines.nodes.antstransform as antstransform
import panpipelines.nodes.atlascreate as atlascreate
import panpipelines.nodes.roi_mean_single as roimean
from panpipelines.utils.util_functions import *
from panpipelines.utils.transformer import *
import glob

def create(name, wf_base_dir,labels_dict,createGraph=True,execution={},LOGGER=None):
    # Create workflow
    pan_workflow = Workflow(name=name, base_dir=wf_base_dir)

    if LOGGER:
        LOGGER.info(f"Created Workflow {name} with base directory {wf_base_dir}")

    if len(execution.keys()) > 0:
        pan_workflow.config = process_dict(pan_workflow.config,execution)

    # do we need to create a custom atlas?
    atlas_index = getParams(labels_dict,"ATLAS_INDEX")
    atlas_file = getParams(labels_dict,"ATLAS_FILE")
    atlas_name = getParams(labels_dict,"ATLAS_NAME")
    atlascreate_node=None
    if not atlas_file:
        newatlas_list=[]
        newatlas_templates = getParams(labels_dict,"NEWATLAS_TEMPLATE")
        newatlas_index = getParams(labels_dict,"NEWATLAS_INDEX")
        if isinstance(newatlas_templates,list):
            for newatlas_template in newatlas_templates:
                evaluated_newatlas_template = substitute_labels(newatlas_template,labels_dict)
                if "*" not in evaluated_newatlas_template:
                    newatlas_list.append(evaluated_newatlas_template)
                else:
                    newatlas_list.extend(glob.glob(evaluated_newatlas_template))
        else:
            newatlas_list.extend(glob.glob(newatlas_templates))
            
        labels_dict = updateParams(labels_dict,"COST_FUNCTION","NearestNeighbor")
        atlascreate_node = atlascreate.create(labels_dict,name=f"atlascreate_{atlas_name}_node",roi_list=newatlas_list,roilabels_list=newatlas_index,LOGGER=LOGGER)

    roimean_node = roimean.create(labels_dict,name="subject_metrics",LOGGER=LOGGER)
    roimean_map_node = MapNode(roimean_node.interface,name="subject_metrics_map",iterfield=['input_file'])

    if atlascreate_node:
        pan_workflow.connect(atlascreate_node,'atlas_index',roimean_map_node,'atlas_index')
    else:
        roimean_map_node.inputs.atlas_index = atlas_index


    atlas_transform_mat=getParams(labels_dict,"ATLAS_TRANSFORM_MAT") 
    atlas_transform_ref=getParams(labels_dict,"ATLAS_TRANSFORM_REF")   
    # should we transform atlas?
    if atlas_transform_mat is not None:
        labels_dict = updateParams(labels_dict,"COST_FUNCTION","NearestNeighbor")
        if atlascreate_node:
            atlas_transform_node = antstransform.create(labels_dict,name="atlas_transform", trans_mat=atlas_transform_mat,ref_file=atlas_transform_ref, LOGGER=LOGGER)
            pan_workflow.connect(atlascreate_node,'atlas_file',atlas_transform_node,'input_file')    
        else:
            atlas_transform_node = antstransform.create(labels_dict,name="atlas_transform",input_file=atlas_file, trans_mat=atlas_transform_mat,ref_file=atlas_transform_ref, LOGGER=LOGGER)
        pan_workflow.connect(atlas_transform_node,'out_file',roimean_map_node,'atlas_file')
    else:
        if atlascreate_node:
            pan_workflow.connect(atlascreate_node,'atlas_file',roimean_map_node,'atlas_file') 
        else:
            roimean_map_node.inputs.atlas_file = atlas_file

    measures_list=[]
    measures_template = getParams(labels_dict,"MEASURES_TEMPLATE")
    if isinstance(measures_template,list):
        for meas_template in measures_template:
            evaluated_meas_template = substitute_labels(meas_template,labels_dict)
            measures_list.extend(glob.glob(evaluated_meas_template))
    else:
        evaluated_measures_template = substitute_labels(measures_template,labels_dict)
        measures_list.extend(glob.glob(evaluated_measures_template))
        
    measures_list.sort()

    # should we transform measures?
    measures_transform_mat=getParams(labels_dict,"MEASURES_TRANSFORM_MAT")
    measures_transform_ref=getParams(labels_dict,"MEASURES_TRANSFORM_REF")
    if measures_transform_mat is not None:
        # go back to default cost function lancsoz
        labels_dict = removeParam(labels_dict,"COST_FUNCTION")
        labels_dict = removeParam(labels_dict,"OUTPUT_TYPE")
        measures_transform_node = antstransform.create(labels_dict,name="measure_transform",trans_mat=measures_transform_mat,ref_file=measures_transform_ref,LOGGER=LOGGER)
        measures_transform_map_node = MapNode(measures_transform_node.interface,name="measure_transform_map",iterfield=['input_file'])
        measures_transform_map_node.inputs.input_file = measures_list
        pan_workflow.connect(measures_transform_map_node,'out_file',roimean_map_node,'input_file')
            
    else:
        roimean_map_node.inputs.input_file = measures_list


    if createGraph:
         pan_workflow.write_graph(graph2use='flat')


    return pan_workflow



from sdcflows import fieldmaps as sfm
from nipype import Workflow, Node
from nipype.interfaces.io import DataSink
from nipype import logging as nlogging
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from pathlib import Path
from functools import partial
import os

IFLOGGER=nlogging.getLogger('nipype.interface')


def parse_params():
    parser = ArgumentParser(description="SDCflows fieldmap workflow")
    parser.add_argument("--fmap_sources", nargs="+", help="Fmap sources")
    parser.add_argument("--subject", help="Subject")
    parser.add_argument("--session", help="Session")
    parser.add_argument("--fieldmap_dir", type=Path, help="Fieldmap directory")
    parser.add_argument("--workdir", type=Path, help="Workflow directort")
    return parser

def sdcflows_phdiff_fieldmap(fmap_sources, subject, session,fieldmap_dir,workdir):
    estimator = sfm.FieldmapEstimation (
        sources = fmap_sources
    )

    phdiff_wf = estimator.get_workflow()

    fm_wf = Workflow(name = "sdcflows_phdiff_wf", base_dir=workdir)
    sinker = Node(DataSink(),name='fmap_sink')
    sinker.inputs.base_directory = os.path.dirname(fieldmap_dir)
    sinker.inputs.substitutions = [
        ('clipped',f"sub-{subject}_ses-{session}_desc-magnitude_fieldmap"),
        ('phase1_rads_phdiff_unwrapped_fmap_extra',f"desc-preproc_fieldmap")
    ]
    basename = os.path.basename(fieldmap_dir)
    fm_wf.connect(phdiff_wf,"outputnode.fmap",sinker,f"{basename}")
    fm_wf.connect(phdiff_wf,"outputnode.fmap_coeff",sinker,f"{basename}.@coeff")
    fm_wf.connect(phdiff_wf,"outputnode.fmap_mask",sinker,f"{basename}.@mask")
    fm_wf.connect(phdiff_wf,"outputnode.fmap_ref",sinker,f"{basename}.@ref")

    fm_wf.run()

def main():
    parser=parse_params()
    args, unknown_args = parser.parse_known_args()
    fmap_sources = args.fmap_sources
    subject = args.subject
    session = args.session
    fieldmap_dir = str(args.fieldmap_dir)
    workdir = str(args.workdir)

    sdcflows_phdiff_fieldmap(fmap_sources, subject, session,fieldmap_dir,workdir)

# This is the standard boilerplate that calls the main() function.
if __name__ == '__main__':
    main()





from panpipelines import Factory
from panpipelines.pipelines.aslprep_panpipeline import aslprep_panpipeline

def test_PANFactory_getpipeline() -> None:
    panFactory = Factory.getPANFactory()
    assert panFactory.get_processflow("aslprep_panpipeline") == aslprep_panpipeline


from panpipelines.utils.util_functions import *
from panpipelines.pipelines.panpipeline import *
import os
import glob
import panpipelines.workflows.volmeasures_workflow as panworkflow

TARGET_ANALYSIS_LEVEL=["participant"]
class volmeasures_panpipeline(panpipeline):

    def __init__(self,labels_dict,pipeline_dir, participant_label, name='volmeasures_panpipeline',createGraph=True,LOGGER=None,execution={},analysis_level = "participant", participant_project=None,participant_session=None):

        if analysis_level not in TARGET_ANALYSIS_LEVEL:
            if LOGGER:
                LOGGER.error(f"pipeline {name} not defined for {analysis_level}")
                raise ValueError(f"pipeline {name} not defined for {analysis_level}")

        super().__init__(labels_dict,pipeline_dir, participant_label,name,createGraph,LOGGER,execution,analysis_level, participant_project,participant_session)

    def proc(self):
        workflow_dir = self.pipeline_dir
        workflow_name = "{}_wf".format(self.name)

        pan_workflow = panworkflow.create(workflow_name,workflow_dir,self.labels_dict,createGraph=self.createGraph,execution=self.execution, LOGGER=self.LOGGER)
        pan_workflow.run()








from panpipelines.utils.util_functions import *
from panpipelines.scripts.panscript import *
import os
import glob
from bids import BIDSLayout
import json
import nibabel
import pandas as pd

class aslprep_panscript(panscript):

    def __init__(self,labels_dict,name='fmriprep_panscript',params="",command="",execution={}):
        super().__init__(self,labels_dict,name=name,params=params,command=command)

        self.params = "--participant_label <PARTICIPANT_LABEL>" \
            " --low-mem"\
            " --skip-bids-validation"\
            " --stop-on-first-crash" \
            " --use-syn-sdc"\
            " --fs-license-file <FSLICENSE>"\
            " --ignore fieldmaps"\
            " -w <OUTPUT_DIR>/aslprep_work"

        self.command = "singularity run --cleanenv --nv --no-home <ASLPREP_CONTAINER>"\
                " <BIDS_DIR>"\
                " <OUTPUT_DIR>/aslprep_output"\
                " participant"

        self.asl_bids_changes = {
            "acq-prod" : {
                "BolusCutOffDelayTechnique": "QUIPSSII",
                "BolusCutOffTechnique": "QUIPSSII",
                "M0Type" : "Separate",
                "RepetitionTimePreparation": { 
                    "choices" : [["field","RepetitionTimeExcitation"],["field","RepetitionTime"],["float",4]]
                }
            },
            "acq-pcasl" : {
                "M0Type" : "Separate",
                "RepetitionTimePreparation": {
                    "choices" : [["field","RepetitionTimeExcitation"],["field","RepetitionTime"],["float",4.2]]
                }
            },
        }

        self.m0_bids_changes = {
            "acq-prod" : {
                "RepetitionTimePreparation": { 
                    "choices" : [["field","RepetitionTimeExcitation"],["field","RepetitionTime"],["float",4]]
                }
            },
            "acq-pcasl" : {
                "RepetitionTimePreparation": {
                    "choices" : [["field","RepetitionTimeExcitation"],["field","RepetitionTime"],["float",10]]
                }
            },
        }

        self.aslcontext_dict = {
            "acq-prod" : "control:label",
            "acq-pcasl" : "label:control"
        }

    def pre_run(self):
        print("pre run - setting template flow directory")
        TEMPLATEFLOW_HOME=getParams(self.labels_dict,"TEMPLATEFLOW_HOME")
        os.environ["TEMPLATEFLOW_HOME"]=TEMPLATEFLOW_HOME
        os.environ["SINGULARITYENV_TEMPLATEFLOW_HOME"]=TEMPLATEFLOW_HOME

        # change fsl out type - important as some aslprep nodes are expecting .nii.gz explictly
        os.environ["FSLOUTPUTTYPE"]="NIFTI_GZ"
        os.environ["SINGULARITYENV_FSLOUTPUTTYPE"]="NIFTI_GZ"
        #Environment variable SINGULARITYENV_FSLOUTPUTTYPE= is set, but APPTAINERENV_FSLOUTPUTTYPE= is preferred

        print("Creating output directory")
        output_dir = getParams(self.labels_dict,"OUTPUT_DIR")
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        bids_dir = getParams(self.labels_dict,"BIDS_DIR")
        participant_label = getParams(self.labels_dict,"PARTICIPANT_LABEL")

        layout = BIDSLayout(bids_dir)
        asl=layout.get(subject=participant_label,suffix='asl', extension='nii.gz')

        if len(asl) > 0:
            asl_entities = asl[0].get_entities()
            if "acquisition" in asl_entities.keys():
                asl_acq = "acq-" + asl_entities["acquisition"]
            else:
                asl_acq = get_bidstag("acq",asl[0].filename)

            # Edit metadata 
            asl_assocs = asl[0].get_associations()
            asl_json_choice = [x.path for x in asl[0].get_associations() if x.entities['extension']==".json"]
            if not asl_json_choice  is None and len(asl_json_choice ) > 0:
                asl_json_file=asl_json_choice[0]
            else:
                asl_entities['extension']=".json"
                asl_json_choice  = layout.get(return_type='file', invalid_filters='allow', **asl_entities)
                if not asl_json_choice  is None and len(asl_json_choice ) > 0:
                    asl_json_file=asl_json_choice[0]

            if not asl_json_file is None and len(asl_json_file) > 0:
                with open(asl_json_file, 'r') as infile:
                    asl_json = json.load(infile) 

                asl_bids_amendments = self.asl_bids_changes[asl_acq]
                for itemkey, itemvalue in asl_bids_amendments.items():
                    if isinstance(itemvalue,dict):
                        if "choices" in itemvalue.keys():
                            choice_list = itemvalue["choices"]
                            for choice in choice_list:
                                field_type=choice[0]
                                field_name=choice[1]
                                if field_type == "field":
                                    if field_name in asl_json.keys():
                                        asl_json[itemkey]=asl_json[field_name]
                                        break
                                else:
                                    varvalue = get_value_bytype(field_type,field_name)
                                    if varvalue is not None:
                                        asl_json[itemkey]=varvalue

                        else:
                            asl_json[itemkey] = itemvalue
                    else:
                        asl_json[itemkey] = itemvalue


                with open(asl_json_file, 'w') as outfile:
                    json.dump(asl_json, outfile ,indent=2)

            else:
                print("No asl json file found.")

            # create aslcontext file
            asl_entities['extension']=".tsv"
            asl_entities['suffix']="aslcontext"
            aslcontext_filepath = layout.build_path(asl_entities)

            asl_img = nibabel.load(asl[0])
            asl_volumes = asl_img.shape[3]
            aslcontext_order = self.aslcontext_dict[asl_acq]
            aslcontext_header = ["volume_type"]
            aslcontext_values = []
            if aslcontext_order == "control:label":
                for count in range(asl_volumes):
                    if count%2 == 0:
                        aslcontext_values.append("control")
                    else:
                        aslcontext_values.append("label")
            elif aslcontext_order == "label:control":
                for count in range(asl_volumes):
                    if count%2 == 0:
                        aslcontext_values.append("label")
                    else:
                        aslcontext_values.append("control")

            aslcontext_df = pd.DataFrame(aslcontext_values,columns = aslcontext_header)
            aslcontext_df.to_csv(aslcontext_filepath,sep="\t",header=True, index=False)

            #m0_choice = [x.path for x in asl[0].get_associations() if x.entities['suffix']=="m0scan"]

        else:
            print("No asl file found.")




    
    def post_run(self):
        print("post run")

    def get_results(self):
        self.results = {}
        return self.results





